import logging
import anyio

from moat.mqtt.client import open_mqttclient
from moat.mqtt.mqtt.constants import QOS_1, QOS_2


#
# This sample shows how to publish messages to broker using different QOS
# Debug outputs shows the message flows
#

logger = logging.getLogger(__name__)

config = {
    "will": {
        "topic": "/will/client",
        "message": b"Dead or alive",
        "qos": 0x01,
        "retain": True,
    }
}


async def test_coro():
    async with open_mqttclient(config=config) as C:
        await C.connect("mqtts://test.mosquitto.org/", cafile="mosquitto.org.crt")
        async with anyio.create_task_group() as tg:
            tg.start_soon(C.publish("a/b", b"TEST MESSAGE WITH QOS_0"))
            tg.start_soon(C.publish("a/b", b"TEST MESSAGE WITH QOS_1", qos=QOS_1))
            tg.start_soon(C.publish("a/b", b"TEST MESSAGE WITH QOS_2", qos=QOS_2))
        logger.info("messages published")


if __name__ == "__main__":
    formatter = "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.DEBUG, format=formatter)
    anyio.run(test_coro)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.

from functools import partial


class EventLoggerPlugin:
    def __init__(self, context):
        self.context = context

    async def log_event(self, *args, **kwargs):  # pylint: disable=unused-argument
        self.context.logger.info(
            "### '%s' EVENT FIRED ###", kwargs["event_name"].replace("old", "")
        )

    def __getattr__(self, name):
        if name.startswith("on_"):
            return partial(self.log_event, event_name=name)


class PacketLoggerPlugin:
    def __init__(self, context):
        self.context = context

    async def on_mqtt_packet_received(self, *args, **kwargs):  # pylint: disable=unused-argument
        packet = kwargs.get("packet")
        session = kwargs.get("session", None)
        if session:
            self.context.logger.debug("%s <-in-- %r", session.client_id, packet)
        else:
            self.context.logger.debug("<-in-- %r", packet)

    async def on_mqtt_packet_sent(self, *args, **kwargs):  # pylint: disable=unused-argument
        packet = kwargs.get("packet")
        session = kwargs.get("session", None)
        if session:
            self.context.logger.debug("%s -out-> %r", session.client_id, packet)
        else:
            self.context.logger.debug("-out-> %r", packet)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import logging

import anyio
from moat.util import create_queue

from ...adapters import StreamAdapter
from ...errors import MQTTException, NoDataException
from ...plugins.manager import PluginManager
from ...session import Session
from ...utils import format_client_message
from ..connack import (
    BAD_USERNAME_PASSWORD,
    CONNECTION_ACCEPTED,
    IDENTIFIER_REJECTED,
    NOT_AUTHORIZED,
    UNACCEPTABLE_PROTOCOL_VERSION,
    ConnackPacket,
)
from ..connect import ConnectPacket
from ..pingreq import PingReqPacket
from ..pingresp import PingRespPacket
from ..suback import SubackPacket
from ..subscribe import SubscribePacket
from ..unsuback import UnsubackPacket
from ..unsubscribe import UnsubscribePacket
from .handler import EVENT_MQTT_PACKET_RECEIVED, EVENT_MQTT_PACKET_SENT, ProtocolHandler

logger = logging.getLogger(__name__)


class BrokerProtocolHandler(ProtocolHandler):
    clean_disconnect = False

    def __init__(self, plugins_manager: PluginManager, session: Session = None):
        super().__init__(plugins_manager, session)
        self._pending_subscriptions = create_queue(9999)
        self._pending_unsubscriptions = create_queue(9999)

    async def handle_write_timeout(self):
        pass

    async def handle_read_timeout(self):
        raise TimeoutError

    async def _handle_disconnect(self, disconnect, wait=True):  # pylint: disable=arguments-differ
        self.logger.debug("Client disconnecting")
        self.clean_disconnect = False  # depending on 'disconnect' (if set)
        with anyio.fail_after(2, shield=True):
            if wait:
                with anyio.move_on_after(min(self.session.keep_alive, 1)):
                    await self._reader_stopped.wait()
            await self.stop()

    async def handle_connection_closed(self):
        if not self._disconnecting:
            self._disconnecting = True
            await self._handle_disconnect(None, wait=False)

    async def handle_connect(self, connect: ConnectPacket):
        # Broker handler shouldn't received CONNECT message during messages handling
        # as CONNECT messages are managed by the broker on client connection
        self.logger.error(
            "%s [MQTT-3.1.0-2] %s : CONNECT message received during messages handling",
            self.session.client_id,
            format_client_message(self.session),
        )
        await self.stop()

    async def handle_pingreq(self, pingreq: PingReqPacket):
        await self._send_packet(PingRespPacket.build())

    async def handle_subscribe(self, subscribe: SubscribePacket):
        subscription = {
            "packet_id": subscribe.variable_header.packet_id,
            "topics": subscribe.payload.topics,
        }
        await self._pending_subscriptions.put(subscription)

    async def handle_unsubscribe(
        self, unsubscribe: UnsubscribePacket
    ):  # pylint: disable=arguments-differ
        unsubscription = {
            "packet_id": unsubscribe.variable_header.packet_id,
            "topics": unsubscribe.payload.topics,
        }
        await self._pending_unsubscriptions.put(unsubscription)

    async def get_next_pending_subscription(self):
        subscription = await self._pending_subscriptions.get()
        return subscription

    async def get_next_pending_unsubscription(self):
        unsubscription = await self._pending_unsubscriptions.get()
        return unsubscription

    async def mqtt_acknowledge_subscription(self, packet_id, return_codes):
        suback = SubackPacket.build(packet_id, return_codes)
        await self._send_packet(suback)

    async def mqtt_acknowledge_unsubscription(self, packet_id):
        unsuback = UnsubackPacket.build(packet_id)
        await self._send_packet(unsuback)

    async def mqtt_connack_authorize(self, authorize: bool):
        if authorize:
            connack = ConnackPacket.build(self.session.parent, CONNECTION_ACCEPTED)
        else:
            connack = ConnackPacket.build(self.session.parent, NOT_AUTHORIZED)
        await self._send_packet(connack)

    @classmethod
    async def init_from_connect(cls, stream: StreamAdapter, plugins_manager):
        """

        :param stream:
        :param plugins_manager:
        :return:
        """
        remote_address, remote_port = stream.get_peer_info()
        try:
            connect = await ConnectPacket.from_stream(stream)
        except NoDataException:
            raise MQTTException("Client closed the connection")  # pylint:disable=W0707
        logger.debug("< B %r", connect)
        await plugins_manager.fire_event(EVENT_MQTT_PACKET_RECEIVED, packet=connect)
        # this shouldn't be required anymore since broker generates for each client a random client_id if not provided
        # [MQTT-3.1.3-6]
        if connect.payload.client_id is None:
            raise MQTTException("[[MQTT-3.1.3-3]] : Client identifier must be present")

        if connect.variable_header.will_flag:
            if connect.payload.will_topic is None or connect.payload.will_message is None:
                raise MQTTException("will flag set, but will topic/message not present in payload")

        if connect.variable_header.reserved_flag:
            raise MQTTException("[MQTT-3.1.2-3] CONNECT reserved flag must be set to 0")
        if connect.proto_name != "MQTT":
            raise MQTTException(
                '[MQTT-3.1.2-1] Incorrect protocol name: "%s"' % connect.proto_name
            )

        connack = None
        error_msg = None
        if connect.proto_level != 4:
            # only MQTT 3.1.1 supported
            error_msg = "Invalid protocol from %s: %d" % (
                format_client_message(address=remote_address, port=remote_port),
                connect.proto_level,
            )
            connack = ConnackPacket.build(
                0, UNACCEPTABLE_PROTOCOL_VERSION
            )  # [MQTT-3.2.2-4] session_parent=0
        elif not connect.username_flag and connect.password_flag:
            connack = ConnackPacket.build(0, BAD_USERNAME_PASSWORD)  # [MQTT-3.1.2-22]
        elif connect.username_flag and not connect.password_flag:
            connack = ConnackPacket.build(0, BAD_USERNAME_PASSWORD)  # [MQTT-3.1.2-22]
        elif connect.username_flag and connect.username is None:
            error_msg = "Invalid username from %s" % (
                format_client_message(address=remote_address, port=remote_port)
            )
            connack = ConnackPacket.build(
                0, BAD_USERNAME_PASSWORD
            )  # [MQTT-3.2.2-4] session_parent=0
        elif connect.password_flag and connect.password is None:
            error_msg = "Invalid password %s" % (
                format_client_message(address=remote_address, port=remote_port)
            )
            connack = ConnackPacket.build(
                0, BAD_USERNAME_PASSWORD
            )  # [MQTT-3.2.2-4] session_parent=0
        elif connect.clean_session_flag is False and (connect.payload.client_id_is_random):
            error_msg = (
                "[MQTT-3.1.3-8] [MQTT-3.1.3-9] %s: No client Id provided (cleansession=0)"
                % (format_client_message(address=remote_address, port=remote_port))
            )
            connack = ConnackPacket.build(0, IDENTIFIER_REJECTED)
        if connack is not None:
            logger.debug("B > %r", connack)
            await plugins_manager.fire_event(EVENT_MQTT_PACKET_SENT, packet=connack)
            await connack.to_stream(stream)

            await stream.close()
            raise MQTTException(error_msg)

        incoming_session = Session(plugins_manager)
        incoming_session.client_id = connect.client_id
        incoming_session.clean_session = connect.clean_session_flag
        incoming_session.will_flag = connect.will_flag
        incoming_session.will_retain = connect.will_retain_flag
        incoming_session.will_qos = connect.will_qos
        incoming_session.will_topic = connect.will_topic
        incoming_session.will_message = connect.will_message
        incoming_session.username = connect.username
        incoming_session.password = connect.password
        if connect.keep_alive > 0:
            incoming_session.keep_alive = connect.keep_alive
        else:
            incoming_session.keep_alive = 0

        handler = cls(plugins_manager)
        return handler, incoming_session



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import logging
import unittest

import anyio
import pytest

from moat.mqtt.plugins.manager import PluginManager

from .. import anyio_run

formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
logging.basicConfig(level=logging.INFO, format=formatter)

pytestmark = pytest.mark.skip


class SimpleTestPlugin:
    def __init__(self, context):
        self.context = context


class EventTestPlugin:
    def __init__(self, context):
        self.context = context
        self.test_flag = False
        self.coro_flag = False

    async def on_test(self, *args, **kwargs):  # pylint: disable=unused-argument
        self.test_flag = True
        self.context.logger.info("on_test")

    async def test_coro(self, *args, **kwargs):  # pylint: disable=unused-argument
        self.coro_flag = True

    async def ret_coro(self, *args, **kwargs):  # pylint: disable=unused-argument
        return "TEST"


class TestPluginManager(unittest.TestCase):
    def test_load_plugin(self):
        async def coro():
            async with anyio.create_task_group() as tg:
                manager = PluginManager(tg, "moat.mqtt.test.plugins", context=None)
                self.assertTrue(len(manager._plugins) > 0)

        anyio_run(coro)

    def test_fire_event(self):
        async def fire_event(manager):
            await manager.fire_event("test")
            await anyio.sleep(1)
            await manager.close()

        async def coro():
            async with anyio.create_task_group() as tg:
                manager = PluginManager(tg, "moat.mqtt.test.plugins", context=None)
                await fire_event(manager)
                plugin = manager.get_plugin("event_plugin")
                self.assertTrue(plugin.object.test_flag)

        anyio_run(coro)

    def test_fire_event_wait(self):
        async def fire_event(manager):
            await manager.fire_event("test", wait=True)
            await manager.close()

        async def coro():
            async with anyio.create_task_group() as tg:
                manager = PluginManager(tg, "moat.mqtt.test.plugins", context=None)
                await fire_event(manager)
                plugin = manager.get_plugin("event_plugin")
                self.assertTrue(plugin.object.test_flag)

        anyio_run(coro)

    def test_map_coro(self):
        async def call_coro(manager):
            await manager.map_plugin_coro("test_coro")

        async def coro():
            async with anyio.create_task_group() as tg:
                manager = PluginManager(tg, "moat.mqtt.test.plugins", context=None)
                await call_coro(manager)
                plugin = manager.get_plugin("event_plugin")
                self.assertTrue(plugin.object.test_coro)

        anyio_run(coro)

    def test_map_coro_return(self):
        async def call_coro(manager):
            return await manager.map_plugin_coro("ret_coro")

        async def coro():
            async with anyio.create_task_group() as tg:
                manager = PluginManager(tg, "moat.mqtt.test.plugins", context=None)
                ret = await call_coro(manager)
                plugin = manager.get_plugin("event_plugin")
                self.assertEqual(ret[plugin], "TEST")

        anyio_run(coro)

    def test_map_coro_filter(self):
        """
        Run plugin coro but expect no return as an empty filter is given
        :return:
        """

        async def call_coro(manager):
            return await manager.map_plugin_coro("ret_coro", filter_plugins=[])

        async def coro():
            async with anyio.create_task_group() as tg:
                manager = PluginManager(tg, "moat.mqtt.test.plugins", context=None)
                ret = await call_coro(manager)
                self.assertTrue(len(ret) == 0)

        anyio_run(coro)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import unittest

from moat.mqtt.adapters import BufferAdapter
from moat.mqtt.errors import MQTTException
from moat.mqtt.mqtt.packet import CONNECT, MQTTFixedHeader

from .. import anyio_run


class TestMQTTFixedHeaderTest(unittest.TestCase):
    def test_from_bytes(self):
        data = b"\x10\x7f"
        stream = BufferAdapter(data)
        header = anyio_run(MQTTFixedHeader.from_stream, stream)
        self.assertEqual(header.packet_type, CONNECT)
        self.assertFalse(header.flags & 0x08)
        self.assertEqual((header.flags & 0x06) >> 1, 0)
        self.assertFalse(header.flags & 0x01)
        self.assertEqual(header.remaining_length, 127)

    def test_from_bytes_with_length(self):
        data = b"\x10\xff\xff\xff\x7f"
        stream = BufferAdapter(data)
        header = anyio_run(MQTTFixedHeader.from_stream, stream)
        self.assertEqual(header.packet_type, CONNECT)
        self.assertFalse(header.flags & 0x08)
        self.assertEqual((header.flags & 0x06) >> 1, 0)
        self.assertFalse(header.flags & 0x01)
        self.assertEqual(header.remaining_length, 268435455)

    def test_from_bytes_ko_with_length(self):
        data = b"\x10\xff\xff\xff\xff\x7f"
        stream = BufferAdapter(data)
        with self.assertRaises(MQTTException):
            anyio_run(MQTTFixedHeader.from_stream, stream)

    def test_to_bytes(self):
        header = MQTTFixedHeader(CONNECT, 0x00, 0)
        data = header.to_bytes()
        self.assertEqual(data, b"\x10\x00")

    def test_to_bytes_2(self):
        header = MQTTFixedHeader(CONNECT, 0x00, 268435455)
        data = header.to_bytes()
        self.assertEqual(data, b"\x10\xff\xff\xff\x7f")



import logging
import anyio

from moat.mqtt.client import open_mqttclient, ClientException
from moat.mqtt.mqtt.constants import QOS_1


#
# This sample shows how to subscbribe a topic and receive data from incoming messages
# It subscribes to '$SYS/broker/uptime' topic and displays the first ten values returned
# by the broker.
#

logger = logging.getLogger(__name__)


async def uptime_coro():
    async with open_mqttclient() as C:
        await C.connect("mqtt://test:test@0.0.0.0:1883")
        # Subscribe to '$SYS/broker/uptime' with QOS=1
        try:
            await C.subscribe(
                [
                    ("data/memes", QOS_1),  # Topic allowed
                    ("data/classified", QOS_1),  # Topic forbidden
                    ("repositories/mqtt/master", QOS_1),  # Topic allowed
                    ("repositories/mqtt/devel", QOS_1),  # Topic forbidden
                    ("calendar/mqtt/releases", QOS_1),  # Topic allowed
                ]
            )
            logger.info("Subscribed")
            for i in range(1, 100):
                message = await C.deliver_message()
                packet = message.publish_packet
                print(
                    "%d: %s => %s"
                    % (i, packet.variable_header.topic_name, str(packet.payload.data))
                )
            await C.unsubscribe(["$SYS/broker/uptime", "$SYS/broker/load/#"])
            logger.info("UnSubscribed")
        except ClientException as ce:
            logger.error("Client exception: %r", ce)


if __name__ == "__main__":
    formatter = "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=formatter)
    anyio.run(uptime_coro)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import unittest

from moat.mqtt.adapters import BufferAdapter
from moat.mqtt.mqtt.pubrel import PacketIdVariableHeader, PubrelPacket

from .. import anyio_run


class PubrelPacketTest(unittest.TestCase):
    def test_from_stream(self):
        data = b"\x60\x02\x00\x0a"
        stream = BufferAdapter(data)
        message = anyio_run(PubrelPacket.from_stream, stream)
        self.assertEqual(message.variable_header.packet_id, 10)

    def test_to_bytes(self):
        variable_header = PacketIdVariableHeader(10)
        publish = PubrelPacket(variable_header=variable_header)
        out = publish.to_bytes()
        self.assertEqual(out, b"\x62\x02\x00\x0a")



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.

__all__ = ["BaseContext", "PluginManager"]

import copy
import logging
from collections import namedtuple

import anyio
import pkg_resources

try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager

from functools import partial

Plugin = namedtuple("Plugin", ["name", "ep", "object"])


class BaseContext:
    def __init__(self):
        self.config = {}  # compat
        self.logger = None


class PluginManager:
    """
    Wraps setuptools Entry point mechanism to provide a basic plugin system.
    Plugins are loaded for a given namespace (group).
    This plugin manager uses coroutines to run plugin call asynchronously in an event queue
    """

    def __init__(self, tg: anyio.abc.TaskGroup, namespace, context):
        self._tg = tg

        self.logger = logging.getLogger(namespace)
        if context is None:
            self.context = BaseContext()
        else:
            self.context = context
        self._plugins = []
        self._load_plugins(namespace)

    @property
    def app_context(self):
        return self.context

    def _load_plugins(self, namespace):
        self.logger.debug("Loading plugins for namespace %s", namespace)
        plugs = self.context.config.get("plugins", None)
        for ep in pkg_resources.iter_entry_points(group=namespace):
            if plugs is not None and ep.name not in plugs:
                continue
            plugin = self._load_plugin(ep)
            self._plugins.append(plugin)
            self.logger.debug(" Plugin %s ready", plugin.ep.name)

    def _load_plugin(self, ep: pkg_resources.EntryPoint):
        try:
            self.logger.debug(" Loading plugin %s", ep)
            plugin = ep.load(require=True)
            self.logger.debug(" Initializing plugin %s", ep)
            plugin_context = copy.copy(self.app_context)
            plugin_context.logger = self.logger.getChild(ep.name)
            obj = plugin(plugin_context)
            return Plugin(ep.name, ep, obj)
        except ImportError as ie:
            self.logger.warning("Plugin %r import failed: %s", ep, ie)
        except pkg_resources.DistributionNotFound as ue:
            self.logger.warning("Plugin %r dependencies resolution failed: %s", ep, ue)
        except pkg_resources.UnknownExtra as ue:
            self.logger.warning("Plugin %r dependencies resolution failed: %s", ep, ue)

    def get_plugin(self, name):
        """
        Get a plugin by its name from the plugins loaded for the current namespace
        :param name:
        :return:
        """
        for p in self._plugins:
            if p.name == name:
                return p
        return None

    async def close(self):
        """
        Free PluginManager resources and cancel pending event methods
        This method call a close() coroutine for each plugin, allowing plugins to close and free resources
        :return:
        """
        await self.map_plugin_coro("close")

    @property
    def plugins(self):
        """
        Get the loaded plugins list
        :return:
        """
        return self._plugins

    async def fire_event(self, event_name, wait=False, *args, **kwargs):  # pylint: disable=W1113
        """
        Fire an event to plugins.
        PluginManager schedule async calls for each plugin on method called "on_" + event_name
        For example, on_connect will be called on event 'connect'
        Method calls are schedule in the async loop. wait parameter must be set to true to wait until all
        mehtods are completed.
        :param event_name:
        :param args:
        :param kwargs:
        :param wait: indicates if fire_event should wait for plugin calls completion (True), or not
        :return:
        """

        @asynccontextmanager
        async def _event_tg(wait):
            if wait:
                async with anyio.create_task_group() as tg:
                    yield tg
            else:
                yield self._tg

        async def _schedule_coro(tg, coro):
            if kwargs:
                coro = partial(coro, **kwargs)
            tg.start_soon(coro, *args)

        async with _event_tg(wait) as tg:
            event_method_name = "on_" + event_name
            for plugin in self._plugins:
                event_method = getattr(plugin.object, event_method_name, None)
                if event_method:
                    try:
                        await _schedule_coro(tg, event_method)

                    except TypeError:
                        self.logger.error(
                            "Method '%s' on plugin '%s' is not a coroutine",
                            event_method_name,
                            plugin.name,
                        )

    async def map(self, coro, *args, **kwargs):
        """
        Schedule a given coroutine call for each plugin.
        The coro called get the Plugin instance as first argument of its method call
        :param coro: coro to call on each plugin
        :param filter_plugins: list of plugin names to filter (only plugin whose name is in filter are called).
        None will call all plugins. [] will call None.
        :param args: arguments to pass to coro
        :param kwargs: arguments to pass to coro
        :return: dict containing return from coro call for each plugin
        """

        async def worker(plugin):
            res = await coro(plugin, *args, **kwargs)
            ret_dict[plugin] = res

        p_list = kwargs.pop("filter_plugins", None)
        if p_list is None:
            p_list = [p.name for p in self.plugins]
        ret_dict = {}
        async with anyio.create_task_group() as tg:
            for plugin in self._plugins:
                if plugin.name in p_list:
                    try:
                        tg.start_soon(worker, plugin)
                    except TypeError:
                        self.logger.error(
                            "Method '%r' on plugin '%s' is not a coroutine",
                            coro,
                            plugin.name,
                        )
        return ret_dict

    @staticmethod
    async def _call_coro(plugin, coro_name, *args, **kwargs):
        try:
            coro = getattr(plugin.object, coro_name, None)(*args, **kwargs)
        except TypeError:
            # Plugin doesn't implement coro_name
            return None
        else:
            return await coro

    async def map_plugin_coro(self, coro_name, *args, **kwargs):
        """
        Call a plugin declared by plugin by its name
        :param coro_name:
        :param args:
        :param kwargs:
        :return:
        """
        return await self.map(self._call_coro, coro_name, *args, **kwargs)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import logging
import os
import random
import unittest

import anyio
import pytest
import trio

try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager

from moat.mqtt.broker import create_broker
from moat.mqtt.client import open_mqttclient
from moat.mqtt.mqtt.constants import QOS_0

try:
    from distkv.client import open_client
    from distkv.server import Server
except ImportError:
    pytestmark = pytest.mark.skip

from functools import partial

from . import anyio_run

formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
logging.basicConfig(level=logging.DEBUG, format=formatter)
log = logging.getLogger(__name__)

# Port for distkv-based broker
PORT = 40000 + (os.getpid() + 5) % 10000
# Port for base broker
PORT_B = 40000 + (os.getpid() + 6) % 10000
# Port for distkv
PORT_D = 40000 + (os.getpid() + 7) % 10000

URI = "mqtt://127.0.0.1:%d/" % PORT
URI_B = "mqtt://127.0.0.1:%d/" % PORT_B

log.debug("Ports: distkv=%d up=%d low=%d", PORT_D, PORT, PORT_B)

broker_config = {
    "broker": {"uri": "mqtt://127.0.0.1:%d" % PORT_B},
    "distkv": {
        "topic": "test_" + "".join(random.choices("abcdefghijklmnopqrstuvwxyz", k=7)),
        "base": ("test", "retain"),
        "transparent": (("test", "vis"),),
        "connect": {"port": PORT_D},
        "server": {
            "backend": "mqtt",
            "mqtt": {"uri": URI_B},
            "bind": [{"host": "localhost", "port": PORT_D, "ssl": False}],
        },
    },
    "listeners": {
        "default": {"type": "tcp", "bind": "127.0.0.1:%d" % (PORT), "max_connections": 10}
    },
    "sys_interval": 0,
    "auth": {"allow-anonymous": True},
}


test_config = {
    "listeners": {
        "default": {"type": "tcp", "bind": "127.0.0.1:%d" % PORT_B, "max_connections": 10}
    },
    "sys_interval": 0,
    "retain": False,
    "auth": {"allow-anonymous": True},
}


@asynccontextmanager
async def distkv_server(n):
    msgs = []
    async with anyio.create_task_group() as tg:
        async with create_broker(test_config, plugin_namespace="moat.mqtt.test.plugins"):
            s = Server("test", cfg=broker_config["distkv"], init="test")
            evt = anyio.Event()
            tg.start_soon(partial(s.serve, ready_evt=evt))
            await evt.wait()

            async with open_client(**broker_config["distkv"]) as cl:

                async def msglog(evt):
                    try:
                        async with cl._stream(
                            "msg_monitor", topic="*"
                        ) as mon:  # , topic=broker_config['distkv']['topic']) as mon:
                            log.info("Monitor Start")
                            evt.set()
                            async for m in mon:
                                log.info("Monitor Msg %r", m.data)
                                msgs.append(m.data)
                    except Exception:
                        log.exception("DEAD")

                evt = anyio.Event()
                await cl.scope.spawn(msglog, evt)
                await evt.wait()
                yield s
                cl.scope.cancel()
            tg.cancel_scope.cancel()
    if len(msgs) != n:
        log.error("MsgCount %d %d", len(msgs), n)
    # assert len(msgs) == n, msgs


class MQTTClientTest(unittest.TestCase):
    def test_deliver(self):
        data = b"data 123 a"

        async def test_coro():
            async with distkv_server(1):
                async with create_broker(broker_config, plugin_namespace="moat.mqtt.test.plugins"):
                    async with open_mqttclient(config=broker_config["broker"]) as client:
                        self.assertIsNotNone(client.session)
                        ret = await client.subscribe([("test_topic", QOS_0)])
                        self.assertEqual(ret[0], QOS_0)
                        async with open_mqttclient(config=broker_config["broker"]) as client_pub:
                            await client_pub.publish("test_topic", data, QOS_0, retain=False)
                        with anyio.fail_after(0.5):
                            message = await client.deliver_message()
                        self.assertIsNotNone(message)
                        self.assertIsNotNone(message.publish_packet)
                        self.assertEqual(message.data, data)
                        pass  # exit client
                    pass  # exit broker
                pass  # exit server
            pass  # exit test

        anyio_run(test_coro, backend="trio")

    def test_deliver_transparent(self):
        data = b"data 123 t"

        async def test_coro():
            async with distkv_server(1):
                async with create_broker(broker_config, plugin_namespace="moat.mqtt.test.plugins"):
                    async with open_mqttclient(config=broker_config["broker"]) as client:
                        self.assertIsNotNone(client.session)
                        ret = await client.subscribe([("test/vis/foo", QOS_0)])
                        self.assertEqual(ret[0], QOS_0)
                        async with open_mqttclient(config=broker_config["broker"]) as client_pub:
                            await client_pub.publish("test/vis/foo", data, QOS_0, retain=False)
                        with anyio.fail_after(0.5):
                            message = await client.deliver_message()
                        self.assertIsNotNone(message)
                        self.assertIsNotNone(message.publish_packet)
                        self.assertEqual(message.data, data)
                        pass  # exit client
                    pass  # exit broker
                pass  # exit server
            pass  # exit test

        anyio_run(test_coro, backend="trio")

    def test_deliver_direct(self):
        data = b"data 123 b"

        async def test_coro():
            async with distkv_server(0):
                async with create_broker(broker_config, plugin_namespace="moat.mqtt.test.plugins"):
                    async with open_mqttclient(config=broker_config["broker"]) as client:
                        self.assertIsNotNone(client.session)
                        ret = await client.subscribe([("test_topic", QOS_0)])
                        self.assertEqual(ret[0], QOS_0)
                        async with open_mqttclient(config=broker_config["broker"]) as client_pub:
                            await client_pub.publish("test_topic", data, QOS_0, retain=False)
                        with anyio.fail_after(0.5):
                            message = await client.deliver_message()
                        self.assertIsNotNone(message)
                        self.assertIsNotNone(message.publish_packet)
                        self.assertEqual(message.data, data)

        anyio_run(test_coro, backend="trio")

    def test_deliver_timeout(self):
        async def test_coro():
            async with distkv_server(0):
                async with create_broker(broker_config, plugin_namespace="moat.mqtt.test.plugins"):
                    async with open_mqttclient(config=broker_config["broker"]) as client:
                        self.assertIsNotNone(client.session)
                        ret = await client.subscribe([("test_topic", QOS_0)])
                        self.assertEqual(ret[0], QOS_0)
                        with self.assertRaises(TimeoutError):
                            with anyio.fail_after(2):
                                await client.deliver_message()

        anyio_run(test_coro, backend="trio")



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.

import copy
import logging
import ssl
from functools import wraps
from typing import Union
from urllib.parse import urlparse, urlunparse

import anyio

try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager

from asyncwebsockets import create_websocket
from moat.util import create_queue
from wsproto.utilities import ProtocolError

from . import codecs
from .adapters import StreamAdapter, WebSocketsAdapter
from .errors import NoDataException
from .mqtt.connack import CLIENT_ERROR, CONNECTION_ACCEPTED
from .mqtt.constants import QOS_0, QOS_1, QOS_2
from .mqtt.protocol.client_handler import ClientProtocolHandler
from .mqtt.protocol.handler import ProtocolHandlerException
from .plugins.manager import BaseContext, PluginManager
from .session import Session
from .utils import match_topic

_defaults = {
    "keep_alive": 10,
    "ping_delay": 1,
    "default_qos": 0,
    "default_retain": False,
    "auto_reconnect": True,
    "reconnect_max_interval": 10,
    "reconnect_retries": 2,
    "codec": "noop",
}

QSIZE = 100

_codecs = {}
for _t in dir(codecs):
    _c = getattr(codecs, _t)
    if isinstance(_c, type) and issubclass(_c, codecs.BaseCodec):
        try:
            _codecs[_c.name] = _c
        except AttributeError:
            pass


def get_codec(codec, fallback=None, config={}):  # pylint: disable=dangerous-default-value
    if codec is None:
        codec = fallback
    if codec is None:
        codec = config["codec"]
    if isinstance(codec, str):
        codec = config.get("codec_params", {}).get("name", codec)
        codec = _codecs[codec]
    if isinstance(codec, type):
        codec = codec(**config.get("codec_params", {}).get(codec.name, {}))
    return codec


_handler_id = 0


class _set(set):
    topic = None
    qos = None


class ClientException(Exception):
    pass


class ConnectException(ClientException):
    def __init__(self, text=None, return_code=CLIENT_ERROR):
        if text is None:
            text = self.__class__.__name__
        super().__init__(text)
        self.return_code = return_code

    pass


class ClientContext(BaseContext):
    """
    ClientContext is used as the context passed to plugins interacting with the client.
    It act as an adapter to client services from plugins
    """

    def __init__(self, config):
        super().__init__()
        self.config = config


base_logger = logging.getLogger(__name__)


def mqtt_connected(func):
    """
    MQTTClient coroutines decorator which will wait until connection before calling the decorated method.
    :param func: coroutine to be called once connected
    :return: coroutine result
    """

    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        if not self._connected_state.is_set():
            base_logger.warning("Client not connected, waiting for it")
            async with anyio.create_task_group() as tg:

                async def wait_no_more():
                    await self._no_more_connections.wait()
                    raise ClientException("Will not reconnect")

                tg.start_soon(wait_no_more)

                await self._connected_state.wait()
                tg.cancel_scope.cancel()
        return await func(self, *args, **kwargs)

    return wrapper


@asynccontextmanager
async def open_mqttclient(uri=None, client_id=None, config={}, codec=None):
    # pylint: disable=dangerous-default-value
    """
    MQTT client implementation.

    MQTTClient instances provides API for connecting to a broker and send/receive messages using the MQTT protocol.

    :param client_id: MQTT client ID to use when connecting to the broker. If none, it will generated randomly by :func:`moat.mqtt.utils.gen_client_id`
    :param config: Client configuration
    :param codec: Codec to default to, the config or "no-op" if not given.
    :return: async context manager returning a class instance

    Example usage::

        async with open_mqttclient(config=dict(uri="mqtt://my-broker.example")) as client:
            # await client.connect("mqtt://my-broker.example")  # alternate use
            await C.subscribe([
                    ('$SYS/broker/uptime', QOS_1),
                    ('$SYS/broker/load/#', QOS_2),
                ])
            async for msg in client:
                packet = message.publish_packet
                print("%d:  %s => %s" % (i, packet.variable_header.topic_name, str(packet.payload.data)))

    """
    async with anyio.create_task_group() as tg:
        C = MQTTClient(tg, client_id, config=config, codec=codec)
        try:
            if uri is not None:
                config["uri"] = uri
            if "uri" in config:
                known_keys = ("uri", "cleansession", "cafile", "capath", "cadata", "extra_headers")
                kwargs = {k: v for k, v in config.items() if k in known_keys}
                await C.connect(**kwargs)
            yield C
        finally:
            with anyio.move_on_after(2, shield=True):
                await C.disconnect()
                tg.cancel_scope.cancel()


class MQTTClient:
    """
    MQTT client implementation.

    MQTTClient instances provides API for connecting to a broker and send/receive messages using the MQTT protocol.

    :param tg: The task group in which to run open-ended subtasks.
    :param client_id: MQTT client ID to use when connecting to the broker. If none, it will generated randomly by :func:`moat.mqtt.utils.gen_client_id`
    :param config: Client configuration
    :param codec: Codec to default to, the config or "no-op" if not given.
    :return: class instance

    You should use :func:`open_mqttclient` to create an instance of
    this class.
    """

    def __init__(self, tg: anyio.abc.TaskGroup, client_id=None, config=None, codec=None):
        self.logger = logging.getLogger(__name__)
        self.config = copy.deepcopy(_defaults)
        if config is not None:
            self.config.update(config)
        if client_id is not None:
            self.client_id = client_id
        else:
            from moat.mqtt.utils import gen_client_id

            self.client_id = gen_client_id()
            self.logger.debug("Using generated client ID : %s", self.client_id)

        self.session = None
        self._tg = tg
        self._handler = None
        self._disconnect_task = None
        self._connected_state = anyio.Event()
        self._no_more_connections = anyio.Event()
        self.extra_headers = {}
        self.codec = get_codec(codec, config=self.config)

        self._subscriptions = None

        # Init plugins manager
        context = ClientContext(self.config)
        self.plugins_manager = PluginManager(tg, "moat.mqtt.client.plugins", context)
        self.client_task = None

    async def connect(
        self,
        uri=None,
        cleansession=None,
        cafile=None,
        capath=None,
        cadata=None,
        extra_headers={},
    ):
        # pylint: disable=dangerous-default-value
        """
        Connect to a remote broker.

        At first, a network connection is established with the server using the given protocol (``mqtt``, ``mqtts``, ``ws`` or ``wss``). Once the socket is connected, a `CONNECT <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718028>`_ message is sent with the requested informations.

        :param uri: Broker URI connection, conforming to `MQTT URI scheme <https://github.com/mqtt/mqtt.github.io/wiki/URI-Scheme>`_. Uses ``uri`` config attribute by default.
        :param cleansession: MQTT CONNECT clean session flag
        :param cafile: server certificate authority file (optional, used for secured connection)
        :param capath: server certificate authority path (optional, used for secured connection)
        :param cadata: server certificate authority data (optional, used for secured connection)
        :param extra_headers: a dictionary with additional http headers that should be sent on the initial connection (optional, used only with websocket connections)
        :return: `CONNACK <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718033>`_ return code
        :raise: :class:`moat.mqtt.client.ConnectException` if connection fails
        """
        self.session = self._initsession(uri, cleansession, cafile, capath, cadata)
        self.extra_headers = extra_headers
        self.logger.debug("Connect to: %s", uri)

        try:
            return await self._do_connect()
        except SyntaxError as be:
            self.logger.warning("Connection failed: %r", be)
            auto_reconnect = self.config.get("auto_reconnect", False)
            if not auto_reconnect:
                raise
            else:
                return await self.reconnect()

    async def disconnect(self):
        """
        Disconnect from the connected broker.

        This method sends a `DISCONNECT <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718090>`_ message and closes the network socket.
        """

        # do not reconnect any more
        self.config["auto_reconnect"] = False
        self.cancel_tasks()

        if self.session is not None and self.session.transitions.is_connected():
            if self._disconnect_task is not None:
                self._disconnect_task.cancel()
            await self._handler.mqtt_disconnect()
            self._connected_state = anyio.Event()  # clear
            await self._handler.stop()
            self.session.transitions.disconnect()
        else:
            self.logger.info("Client session is not currently connected, ignoring Disconnect call")

    def cancel_tasks(self):
        """
        Before disconnection need to cancel all pending tasks
        :return:
        """
        if self.client_task is not None:
            task, self.client_task = self.client_task, None
            task.cancel()

    async def reconnect(self, cleansession=None):
        """
        Reconnect a previously connected broker.

        Reconnection tries to establish a network connection and send a `CONNECT <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718028>`_ message.
        Retries interval and attempts can be controled with the ``reconnect_max_interval`` and ``reconnect_retries`` configuration parameters.

        :param cleansession: clean session flag used in MQTT CONNECT messages sent for reconnections.
        :return: `CONNACK <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718033>`_ return code
        :raise: :class:`moat.mqtt.client.ConnectException` if re-connection fails after max retries.
        """

        if self.session.transitions.is_connected():
            self.logger.warning("Client already connected")
            return CONNECTION_ACCEPTED

        if cleansession:
            self.session.clean_session = cleansession
        self.logger.debug("Reconnecting with session parameters: %r", self.session)
        reconnect_max_interval = self.config.get("reconnect_max_interval", 10)
        reconnect_retries = self.config.get("reconnect_retries", 5)
        nb_attempt = 1
        await anyio.sleep(1)
        while True:
            try:
                self.logger.debug("Reconnect attempt %d ...", nb_attempt)
                return await self._do_connect()
            except SyntaxError as e:  # no you can't reconnect on every exception
                self.logger.warning("Reconnection attempt failed: %r", e)
                if reconnect_retries >= 0 and nb_attempt > reconnect_retries:
                    self.logger.error(
                        "Maximum number of connection attempts reached. Reconnection aborted"
                    )
                    raise ConnectException(  # pylint: disable=W0707
                        "Too many connection attempts failed"
                    )
                exp = 2**nb_attempt
                delay = exp if exp < reconnect_max_interval else reconnect_max_interval
                self.logger.debug("Waiting %d second before next attempt", delay)
                await anyio.sleep(delay)
                nb_attempt += 1

    async def _do_connect(self):
        return_code = await self._connect_coro()
        evt = anyio.Event()
        self._tg.start_soon(self.handle_connection_close, evt)
        await evt.wait()
        return return_code

    @mqtt_connected
    async def ping(self):
        """
        Ping the broker.

        Send a MQTT `PINGREQ <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718081>`_ message for response.
        """

        if self.session.transitions.is_connected():
            await self._handler.mqtt_ping()
        else:
            self.logger.warning(
                "MQTT PING request incompatible with current session state '%s'",
                self.session.transitions.state,
            )

    @mqtt_connected
    async def publish(self, topic, message, qos=None, retain=None, codec=None):
        """
        Publish a message to the broker.

        Send a MQTT `PUBLISH <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718037>`_ message and wait for acknowledgment depending on Quality Of Service

        :param topic: topic name to which message data is published
        :param message: payload message (as bytes) to send.
        :param qos: requested publish quality of service : QOS_0, QOS_1 or QOS_2. Defaults to ``default_qos`` config parameter or QOS_0.
        :param retain: retain flag. Defaults to ``default_retain`` config parameter or False.
        :param codec: Codec to encode the message with. Defaults to the connection's.
        """

        codec = get_codec(codec, self.codec, config=self.config)
        message = codec.encode(message)
        if not isinstance(topic, str):
            topic = "/".join(topic)

        if qos is not None:
            assert qos in (QOS_0, QOS_1, QOS_2)
        else:
            try:
                qos = self.config["topics"][topic]["qos"]
            except KeyError:
                qos = self.config["default_qos"]
        if retain is None:
            try:
                retain = self.config["topics"][topic]["retain"]
            except KeyError:
                retain = self.config["default_retain"]
        return await self._handler.mqtt_publish(topic, message, qos, retain)

    @mqtt_connected
    async def subscribe(self, topics):
        """
        Subscribe to some topics.

        Send a MQTT `SUBSCRIBE <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718063>`_ message and wait for broker acknowledgment.

        :param topics: array of topics pattern to subscribe with associated QoS.
        :return: `SUBACK <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718068>`_ message return code.

        Example of ``topics`` argument expected structure:
        ::

            [
                ('$SYS/broker/uptime', QOS_1),
                ('$SYS/broker/load/#', QOS_2),
            ]
        """
        return await self._handler.mqtt_subscribe(topics, self.session.next_packet_id)

    async def unsubscribe(self, topics):
        """
        Unsubscribe from some topics.

        Send a MQTT `UNSUBSCRIBE <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718072>`_ message and wait for broker `UNSUBACK <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718077>`_ message.

        :param topics: array of topics to unsubscribe from.

        Example of ``topics`` argument expected structure:
        ::

            ['$SYS/broker/uptime', '$SYS/broker/load/#']

        This is a no-op if the client is not connected.
        """
        if not self._connected_state.is_set():
            return
        await self._handler.mqtt_unsubscribe(topics, self.session.next_packet_id)

    def subscription(self, topic, qos=QOS_0, codec=None):
        """
        Iterate over a topic.

            :param topic: topic (including wildcards) to receive from.
            :param qos: minimum QOS to receive with.
            :param codec: Codec to decode incoming mssages with.

            Usage::

                async with client.subscription("foo/bar/#") as sub:
                    async for msg in sub:
                        await process(msg)

        You can use either multiple calls to `subscription`, or manage
        message dispatching yourself with `subscribe` and `deliver_message`.
        Using both methods in parallel is not supported.

        """

        class _Subscription:
            def __init__(self, client, topic: Union[str, tuple], qos: int, codec=None):
                self.client = client
                self.topic = topic if isinstance(topic, str) else "/".join(topic)
                self.qos = qos
                if codec is None:
                    codec = client.codec
                elif isinstance(codec, str):
                    codec = _codecs[codec]()
                self.codec = codec
                self._q = None

                global _handler_id
                _handler_id += 1
                self._id = _handler_id

            def __hash__(self):
                return self._id

            def __eq__(self, other):
                return self._id == getattr(other, "_id", other)

            async def __aenter__(self):
                self._q = create_queue(QSIZE)
                await self.client._subscribe(self)
                return self

            async def __aexit__(self, *tb):
                self._q = None
                try:
                    with anyio.move_on_after(2, shield=True):
                        await self.client._unsubscribe(self)
                except ClientException:
                    pass

            def __aiter__(self):
                return self

            async def __anext__(self):
                if self._q is None:
                    raise RuntimeError("Overflow. Please resubscribe.")
                message = await self._q.get()
                message.data = self.codec.decode(message.publish_packet.data)
                return message

            async def __len__(self):
                """Queue length"""
                return self._q.qsize()

            async def publish(self, topic, message, *a, **kw):
                """
                Publish a message.

                :param topic: topic name to which message data is published
                :param message: payload message (as bytes) to send.
                :param qos: requested publish quality of service : QOS_0, QOS_1 or QOS_2. Defaults to ``default_qos`` config parameter or QOS_0.
                :param retain: retain flag. Defaults to ``default_retain`` config parameter or False.
                :param codec: Codec to encode the message with. Defaults to the subscription's.
                """
                if topic is None:
                    topic = self.topic
                if "codec" not in kw:
                    kw["codec"] = self.codec
                await self.client.publish(topic, message, *a, **kw)

        return _Subscription(self, topic, qos, codec=codec)

    async def _subscribe(self, handler):
        topic = handler.topic
        if self._subscriptions is None:
            self._subscriptions = dict()
            self._tg.start_soon(self._deliver_loop)
        clients = self._subscriptions.get(topic, None)
        if clients is None:
            self._subscriptions[topic] = clients = _set()
            clients.topic = tuple(topic.split("/"))
            clients.qos = handler.qos
            await self.subscribe([(topic, handler.qos)])
        elif clients.qos < handler.qos:
            clients.qos = handler.qos
            await self.subscribe([(topic, handler.qos)])

        clients.add(handler)

    async def _unsubscribe(self, handler):
        topic = handler.topic
        clients = self._subscriptions[topic]
        clients.remove(handler)
        if not clients:
            del self._subscriptions[topic]
            await self.unsubscribe([topic])

    async def _dispatch(self, msg):
        t = msg.topic.split("/")
        for clients in list(self._subscriptions.values()):
            if not match_topic(t, clients.topic):
                continue
            for c in list(clients):
                if c._q is None:
                    continue
                if c._q.qsize() < QSIZE - 1:
                    await c._q.put(msg)
                else:
                    await c._q.put(None)
                    c._q = None

    async def _deliver_loop(self):
        """
        Dispatch incoming messages to subscriptions.
        """
        with anyio.CancelScope() as scope:
            if self.client_task is not None:
                raise RuntimeError("You can't listen in more than one task")
            self.client_task = scope
            try:
                while True:
                    if self.session is None:
                        return
                    msg = await self.session.get_next_message()
                    if msg is None:
                        return
                    await self._dispatch(msg)

            finally:
                self.client_task = None

    @mqtt_connected
    async def deliver_message(self, codec=None):
        """
        Deliver next received message.

        Deliver next message received from the broker. If no message is available, this methods waits until next message arrives or ``timeout`` occurs.

        :return: instance of :class:`moat.mqtt.session.ApplicationMessage` containing received message information flow.
        :raises: :class:`TimeoutError` if timeout occurs before a message is delivered

        :param codec: Codec to decode the message with.

        This method returns ``None`` if it is cancelled by closing the
        connection.
        """
        if codec is None:
            codec = self.codec
        elif isinstance(codec, str):
            codec = _codecs[codec]()

        with anyio.CancelScope() as scope:
            if self.session is None:
                return None
            if self.client_task is not None:
                raise RuntimeError("You can't listen in more than one task")
            self.client_task = scope
            try:
                msg = await self.session.get_next_message()
                msg.data = codec.decode(msg.publish_packet.data)
                return msg

            finally:
                self.client_task = None

    async def _connect_coro(self):
        kwargs = dict()

        # Decode URI attributes
        uri_attributes = urlparse(self.session.broker_uri)
        scheme = uri_attributes.scheme
        secure = True if scheme in ("mqtts", "wss") else False
        self.session.username = (
            self.session.username if self.session.username else uri_attributes.username
        )
        self.session.password = (
            self.session.password if self.session.password else uri_attributes.password
        )
        self.session.remote_address = uri_attributes.hostname
        self.session.remote_port = uri_attributes.port
        if scheme in ("mqtt", "mqtts") and not self.session.remote_port:
            self.session.remote_port = 8883 if scheme == "mqtts" else 1883
        if scheme in ("ws", "wss") and not self.session.remote_port:
            self.session.remote_port = 443 if scheme == "wss" else 80
        if scheme in ("ws", "wss"):
            # Rewrite URI to conform to https://tools.ietf.org/html/rfc6455#section-3
            uri = (
                scheme,
                self.session.remote_address + ":" + str(self.session.remote_port),
                uri_attributes[2],
                uri_attributes[3],
                uri_attributes[4],
                uri_attributes[5],
            )
            self.session.broker_uri = urlunparse(uri)
        # Init protocol handler
        # if not self._handler:
        self._handler = ClientProtocolHandler(self.plugins_manager)

        if secure:
            sc = ssl.create_default_context(
                ssl.Purpose.SERVER_AUTH,
                cafile=self.session.cafile,
                capath=self.session.capath,
                cadata=self.session.cadata,
            )
            if "certfile" in self.config and "keyfile" in self.config:
                sc.load_cert_chain(self.config["certfile"], self.config["keyfile"])
            if "check_hostname" in self.config and isinstance(self.config["check_hostname"], bool):
                sc.check_hostname = self.config["check_hostname"]
            kwargs["ssl_context"] = sc
            kwargs["autostart_tls"] = True

        try:
            adapter = None
            self._connected_state = anyio.Event()  # clear
            # Open connection
            if scheme in ("mqtt", "mqtts"):
                conn = await anyio.connect_tcp(
                    self.session.remote_address, self.session.remote_port
                )
                if kwargs.pop("autostart_tls", False):
                    try:
                        conn = await anyio.streams.tls.TLSStream.wrap(
                            conn, ssl_context=kwargs.pop("ssl_context"), server_side=False
                        )
                    except BaseException:
                        with anyio.move_on_after(1, shield=True):
                            await conn.aclose()
                        raise
                adapter = StreamAdapter(conn)
            elif scheme in ("ws", "wss"):
                if kwargs.pop("autostart_tls", False):
                    kwargs["ssl"] = kwargs.pop("ssl_context")
                websocket = await create_websocket(
                    self.session.broker_uri,
                    subprotocols=["mqtt"],
                    headers=self.extra_headers,
                    **kwargs
                )
                adapter = WebSocketsAdapter(websocket)
            else:
                raise ConnectException("Unknown URL scheme")
            # Start MQTT protocol
            await self._handler.attach(self.session, adapter)
            try:
                return_code = await self._handler.mqtt_connect()
            except NoDataException:
                self.logger.warning("Connection broken by broker")
                raise ConnectException("Connection broken by broker")  # pylint: disable=W0707
            if return_code != CONNECTION_ACCEPTED:
                self.session.transitions.disconnect()
                self.logger.warning("Connection rejected with code '%s'", return_code)
                raise ConnectException("Connection rejected by broker", return_code)

            # Handle MQTT protocol
            await self._handler.start()
            self.session.transitions.connect()
            self._connected_state.set()
            topics = []
            if self._subscriptions is not None:
                for s in self._subscriptions.values():
                    topics.append(("/".join(s.topic), s.qos))
                if topics:
                    await self.subscribe(topics)
            self.logger.debug(
                "connected to %s:%s",
                self.session.remote_address,
                self.session.remote_port,
            )
            return return_code
        except ProtocolError as exc:
            self.logger.warning("connection failed: invalid websocket handshake")
            self.session.transitions.disconnect()
            await adapter.close()
            raise ConnectException("connection failed: invalid websocket handshake") from exc
        except (ProtocolHandlerException, ConnectionError, OSError) as exc:
            self.logger.warning("MQTT connection failed")
            self.session.transitions.disconnect()
            if adapter is not None:
                await adapter.close()
            raise ConnectException from exc
        except BaseException:
            with anyio.move_on_after(1, shield=True):
                if adapter is not None:
                    await adapter.close()
            raise

    async def handle_connection_close(self, evt):
        def cancel_tasks():
            self._no_more_connections.set()
            if self.client_task:
                task, self.client_task = self.client_task, None
                task.cancel()

        with anyio.CancelScope() as scope:
            self._disconnect_task = scope
            evt.set()
            # Wait for disconnection from broker (like connection lost)
            await self._handler.wait_disconnect()
        self.logger.warning("Disconnected from broker")

        # Block client API
        self._connected_state = anyio.Event()  # .clear()

        # stop and clean handler
        await self._handler.stop()
        await self._handler.detach()
        self.session.transitions.disconnect()

        if self.config.get("auto_reconnect", False):
            # Try reconnection
            self.logger.debug("Auto-reconnecting")
            try:
                await self.reconnect()
            except ConnectException:
                # Cancel client pending tasks
                cancel_tasks()
        else:
            # Cancel client pending tasks
            cancel_tasks()

    def _initsession(
        self, uri=None, cleansession=None, cafile=None, capath=None, cadata=None
    ) -> Session:
        # Load config
        broker_conf = self.config.get("broker", dict()).copy()
        if uri:
            broker_conf["uri"] = uri
        if cafile:
            broker_conf["cafile"] = cafile
        elif "cafile" not in broker_conf:
            broker_conf["cafile"] = None
        if capath:
            broker_conf["capath"] = capath
        elif "capath" not in broker_conf:
            broker_conf["capath"] = None
        if cadata:
            broker_conf["cadata"] = cadata
        elif "cadata" not in broker_conf:
            broker_conf["cadata"] = None

        if cleansession is not None:
            broker_conf["cleansession"] = cleansession

        for key in ["uri"]:
            if broker_conf.get(key, None) is None:
                raise ClientException("Missing connection parameter '%s'" % key)

        s = Session(self.plugins_manager)
        s.broker_uri = broker_conf["uri"]
        s.client_id = self.client_id
        s.cafile = broker_conf["cafile"]
        s.capath = broker_conf["capath"]
        s.cadata = broker_conf["cadata"]
        if cleansession is not None:
            s.clean_session = cleansession
        else:
            s.clean_session = self.config.get("cleansession", True)
        s.keep_alive = self.config["keep_alive"] - self.config["ping_delay"]
        if "will" in self.config and self.config["will"]["topic"]:
            s.will_flag = True
            s.will_retain = self.config["will"]["retain"]
            s.will_topic = self.config["will"]["topic"]
            s.will_message = self.config["will"]["message"]
            s.will_qos = self.config["will"]["qos"]
        else:
            s.will_flag = False
            s.will_retain = False
            s.will_topic = None
            s.will_message = None
            s.will_qos = 0
        return s



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.

from ..adapters import StreamAdapter
from ..codecs import bytes_to_int, int_to_bytes, read_or_raise
from ..errors import MoatMQTTException, NoDataException
from .packet import (
    SUBACK,
    MQTTFixedHeader,
    MQTTPacket,
    MQTTPayload,
    MQTTVariableHeader,
    PacketIdVariableHeader,
)


class SubackPayload(MQTTPayload):

    __slots__ = ("return_codes",)

    RETURN_CODE_00 = 0x00
    RETURN_CODE_01 = 0x01
    RETURN_CODE_02 = 0x02
    RETURN_CODE_80 = 0x80

    def __init__(self, return_codes=()):
        super().__init__()
        self.return_codes = return_codes

    def __repr__(self):
        return type(self).__name__ + "(return_codes={0})".format(repr(self.return_codes))

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        out = b""
        for return_code in self.return_codes:
            out += int_to_bytes(return_code, 1)
        return out

    @classmethod
    async def from_stream(
        cls,
        reader: StreamAdapter,
        fixed_header: MQTTFixedHeader,
        variable_header: MQTTVariableHeader,
    ):
        return_codes = []
        bytes_to_read = fixed_header.remaining_length - variable_header.bytes_length
        for _ in range(0, bytes_to_read):
            try:
                return_code_byte = await read_or_raise(reader, 1)
                return_code = bytes_to_int(return_code_byte)
                return_codes.append(return_code)
            except NoDataException:
                break
        return cls(return_codes)


class SubackPacket(MQTTPacket):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = SubackPayload

    def __init__(
        self,
        fixed: MQTTFixedHeader = None,
        variable_header: PacketIdVariableHeader = None,
        payload=None,
    ):
        if fixed is None:
            header = MQTTFixedHeader(SUBACK, 0x00)
        else:
            if fixed.packet_type != SUBACK:
                raise MoatMQTTException(
                    "Invalid fixed packet type %s for SubackPacket init" % fixed.packet_type
                )
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @classmethod
    def build(cls, packet_id, return_codes):
        variable_header = cls.VARIABLE_HEADER(packet_id)
        payload = cls.PAYLOAD(return_codes)
        return cls(variable_header=variable_header, payload=payload)



class BaseTopicPlugin:
    def __init__(self, context):
        self.context = context
        try:
            self.topic_config = self.context.config["topic-check"]
        except KeyError:
            self.context.logger.warning("'topic-check' section not found in context configuration")
            self.topic_config = None

    async def topic_filtering(self, *args, **kwargs):  # pylint: disable=unused-argument
        if not self.topic_config:
            # auth config section not found
            return None
        return True


class TopicTabooPlugin(BaseTopicPlugin):
    def __init__(self, context):
        super().__init__(context)
        self._taboo = ["prohibited", "top-secret", "data/classified"]

    async def topic_filtering(self, *args, **kwargs):
        filter_result = await super().topic_filtering(*args, **kwargs)
        if filter_result:
            session = kwargs.get("session", None)
            topic = kwargs.get("topic", None)
            if session.username and session.username == "admin":
                return True
            if topic and topic in self._taboo:
                return False
            return True
        return filter_result


class TopicAccessControlListPlugin(BaseTopicPlugin):
    @staticmethod
    def topic_ac(topic_requested, topic_allowed):
        req_split = topic_requested.split("/")
        allowed_split = topic_allowed.split("/")
        ret = True
        for i in range(max(len(req_split), len(allowed_split))):
            try:
                a_aux = req_split[i]
                b_aux = allowed_split[i]
            except IndexError:
                ret = False
                break
            if b_aux == "#":
                break
            elif (b_aux == "+") or (b_aux == a_aux):
                continue
            else:
                ret = False
                break
        return ret

    async def topic_filtering(self, *args, **kwargs):
        if not self.topic_config:
            # auth config section not found
            return None
        filter_result = await super().topic_filtering(*args, **kwargs)
        if filter_result:
            session = kwargs.get("session", None)
            req_topic = kwargs.get("topic", None)
            if req_topic:
                username = session.username
                if username is None:
                    username = "anonymous"
                try:
                    allowed_topics = self.topic_config["acl"].get(username, [])
                except KeyError:
                    return True
                else:
                    for allowed_topic in allowed_topics:
                        if self.topic_ac(req_topic, allowed_topic):
                            return True
                    return False
            else:
                return False



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import anyio

from ..codecs import decode_packet_id, decode_string, encode_string, int_to_bytes
from ..errors import MoatMQTTException, MQTTException
from .packet import PUBLISH, MQTTFixedHeader, MQTTPacket, MQTTPayload, MQTTVariableHeader


class PublishVariableHeader(MQTTVariableHeader):

    __slots__ = ("topic_name", "packet_id")

    def __init__(self, topic_name: str, packet_id: int = None):
        super().__init__()
        if "*" in topic_name:
            raise MQTTException(
                "[MQTT-3.3.2-2] Topic name in the PUBLISH Packet MUST NOT contain wildcard characters."
            )
        if not isinstance(topic_name, str):
            topic_name = "/".join(topic_name)
        self.topic_name = topic_name
        self.packet_id = packet_id

    def __repr__(self):
        return type(self).__name__ + "(topic={0}, packet_id={1})".format(
            self.topic_name, self.packet_id
        )

    def to_bytes(self):
        out = bytearray()
        out.extend(encode_string(self.topic_name))
        if self.packet_id is not None:
            out.extend(int_to_bytes(self.packet_id, 2))
        return out

    @classmethod
    async def from_stream(cls, reader: anyio.abc.ByteStream, fixed_header: MQTTFixedHeader):
        topic_name = await decode_string(reader)
        has_qos = (fixed_header.flags >> 1) & 0x03
        if has_qos:
            packet_id = await decode_packet_id(reader)
        else:
            packet_id = None
        return cls(topic_name, packet_id)


class PublishPayload(MQTTPayload):

    __slots__ = ("data",)

    def __init__(self, data: bytes = None):
        super().__init__()
        self.data = data

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        return self.data

    @classmethod
    async def from_stream(
        cls,
        reader: anyio.abc.ByteStream,
        fixed_header: MQTTFixedHeader,
        variable_header: MQTTVariableHeader,
    ):
        data = bytearray()
        data_length = fixed_header.remaining_length - variable_header.bytes_length
        length_read = 0
        while length_read < data_length:
            buffer = await reader.read(data_length - length_read)
            data.extend(buffer)
            length_read = len(data)
        return cls(data)

    def __repr__(self):
        return type(self).__name__ + "(data={0!r})".format(repr(self.data))


class PublishPacket(MQTTPacket):
    VARIABLE_HEADER = PublishVariableHeader
    PAYLOAD = PublishPayload

    DUP_FLAG = 0x08
    RETAIN_FLAG = 0x01
    QOS_FLAG = 0x06

    def __init__(
        self,
        fixed: MQTTFixedHeader = None,
        variable_header: PublishVariableHeader = None,
        payload=None,
    ):
        if fixed is None:
            header = MQTTFixedHeader(PUBLISH, 0x00)
        else:
            if fixed.packet_type != PUBLISH:
                raise MoatMQTTException(
                    "Invalid fixed packet type %s for PublishPacket init" % fixed.packet_type
                )
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    def set_flags(self, dup_flag=False, qos=0, retain_flag=False):
        self.dup_flag = dup_flag
        self.retain_flag = retain_flag
        self.qos = qos

    def _set_header_flag(self, val, mask):
        if val:
            self.fixed_header.flags |= mask
        else:
            self.fixed_header.flags &= ~mask

    def _get_header_flag(self, mask):
        if self.fixed_header.flags & mask:
            return True
        else:
            return False

    @property
    def dup_flag(self) -> bool:
        return self._get_header_flag(self.DUP_FLAG)

    @dup_flag.setter
    def dup_flag(self, val: bool):
        self._set_header_flag(val, self.DUP_FLAG)

    @property
    def retain_flag(self) -> bool:
        return self._get_header_flag(self.RETAIN_FLAG)

    @retain_flag.setter
    def retain_flag(self, val: bool):
        self._set_header_flag(val, self.RETAIN_FLAG)

    @property
    def qos(self):
        return (self.fixed_header.flags & self.QOS_FLAG) >> 1

    @qos.setter
    def qos(self, val: int):
        self.fixed_header.flags &= 0xF9
        self.fixed_header.flags |= val << 1

    @property
    def packet_id(self):
        return self.variable_header.packet_id

    @packet_id.setter
    def packet_id(self, val: int):
        self.variable_header.packet_id = val

    @property
    def data(self):
        return self.payload.data

    @data.setter
    def data(self, data: bytes):
        self.payload.data = data

    @property
    def topic_name(self):
        return self.variable_header.topic_name

    @topic_name.setter
    def topic_name(self, name: str):
        self.variable_header.topic_name = name

    @classmethod
    def build(cls, topic_name: str, message: bytes, packet_id: int, dup_flag, qos, retain):
        v_header = PublishVariableHeader(topic_name, packet_id)
        payload = PublishPayload(message)
        packet = PublishPacket(variable_header=v_header, payload=payload)
        packet.dup_flag = dup_flag
        packet.retain_flag = retain
        packet.qos = qos
        return packet



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
from typing import Optional

import anyio

from .broker import Broker

try:
    from distkv.client import client_scope as distkv_client_scope
    from distkv.errors import ErrorRoot
    from moat.util import NotGiven, PathLongener
except ImportError:
    pass


class DistKVbroker(Broker):
    """
    A Broker that routes messages through DistKV / MQTT.
    """

    __slots__ = (
        "_distkv_broker__client",
        "_distkv_broker__topic",
        "_distkv_broker__tranparent",
        "_distkv_broker__base",
    )

    def __init__(self, tg: anyio.abc.TaskGroup, config=None, plugin_namespace=None):
        self.__client = None
        super().__init__(tg, config=config, plugin_namespace=plugin_namespace)
        cfg = self.config["distkv"]

        def spl(x, from_cfg=True):
            if from_cfg:
                x = cfg[x]
            if isinstance(x, str):
                x = x.split("/")
            return tuple(x)

        self.__topic = spl("topic")
        self.__base = spl("base")
        self.__transparent = [spl(x, False) for x in cfg.get("transparent", ())]

        if self.__topic[: len(self.__base)] == self.__base:
            raise ValueError("'topic' must not start with 'base'")

    async def __read_encap(
        self, client, cfg: dict, evt: Optional[anyio.abc.Event] = None
    ):  # pylint: disable=unused-argument
        """
        Read encapsulated messages from the real server and forward them
        """

        async with self.__client.msg_monitor(self.__topic) as q:
            if evt is not None:
                evt.set()
            async for m in q:
                d = m.data
                sess = d.pop("session", None)
                if sess is not None:
                    sess = self._sessions.get(sess, None)
                    if sess is not None:
                        sess = sess[0]
                await super().broadcast_message(session=sess, **d)

    async def __read_topic(
        self, topic, client, cfg: dict, evt: Optional[anyio.abc.Event] = None
    ):  # pylint: disable=unused-argument
        """
        Read topical messages from the real server and forward them
        """
        async with self.__client.msg_monitor(topic, raw=True) as q:
            if evt is not None:
                evt.set()
            async for m in q:
                d = m.raw
                t = m.topic
                await super().broadcast_message(topic=t, data=d, session=None)

    async def __session(self, cfg: dict, evt: Optional[anyio.abc.Event] = None):
        """
        Connect to the real server, read messages, forward them
        """
        try:
            self.__client = client = await distkv_client_scope(connect=cfg["connect"])
            async with anyio.create_task_group() as tg:

                async def start(p, *a):
                    evt = anyio.Event()
                    tg.start_soon(p, *a, client, cfg, evt)
                    await evt.wait()

                if self.__topic:
                    await start(self.__read_encap)
                for t in self.__transparent:
                    await start(self.__read_topic, t)
                    await start(self.__read_topic, t + ("#",))

                if evt is not None:
                    evt.set()
                # The taskgroup waits for it all to finish, i.e. indefinitely
        finally:
            self.__client = None

    async def __retain_reader(
        self, cfg: dict, evt: Optional[anyio.abc.Event] = None
    ):  # pylint: disable=unused-argument
        """
        Read changes from DistKV and broadcast them
        """

        pl = PathLongener(self.__base)
        err = await ErrorRoot.as_handler(self.__client)
        async with self.__client.watch(self.__base, fetch=True, long_path=False) as w:
            evt.set()
            async for msg in w:
                if "path" not in msg:
                    continue
                pl(msg)
                data = msg.get("value", NotGiven)
                if data is NotGiven:
                    data = b""
                elif not isinstance(data, (bytes, bytearray)):
                    await err.record_error(
                        "moat.mqtt", msg.path, data={"data": data}, message="non-binary data"
                    )
                    return
                else:
                    await super().broadcast_message(
                        session=None,
                        topic="/".join(msg["path"]),
                        data=data,
                        retain=True,
                    )
                await err.record_working("moat.mqtt", msg.path)

    async def start(self):
        cfg = self.config["distkv"]

        await super().start()

        evt = anyio.Event()
        self._tg.start_soon(self.__session, cfg, evt)
        await evt.wait()

        evt = anyio.Event()
        self._tg.start_soon(self.__retain_reader, cfg, evt)
        await evt.wait()

    async def broadcast_message(
        self, session, topic, data, force_qos=None, qos=None, retain=False
    ):
        if isinstance(topic, str):
            ts = tuple(topic.split("/"))
        else:
            ts = tuple(topic)
            topic = "/".join(ts)

        if self.__client is None:
            self.logger.error("No client, dropping %s", topic)
            return  # can't do anything

        if topic[0] == "$":
            # $SYS and whatever-else-dollar messages are not DistKV's problem.
            await super().broadcast_message(session, topic, data, retain=retain)
            return

        if ts[: len(self.__base)] == self.__base:
            # All messages on "base" get stored in DistKV, retained or not.
            await self.__client.set(ts, value=data)
            return

        for t in self.__transparent:
            # Messages to be forwarded transparently. The "retain" flag is ignored.
            if len(ts) >= len(t) and t == ts[: len(t)]:
                await self.__client.msg_send(topic=ts, raw=data)
                return

        if self.__topic:
            # Anything else is encapsulated
            msg = dict(session=session.client_id, topic=topic, data=data, retain=retain)
            if qos is not None:
                msg["qos"] = qos
            if force_qos is not None:
                msg["force_qos"] = qos
            await self.__client.msg_send(topic=self.__topic, data=msg)
            return

        self.logger.info("Message ignored: %s %r", topic, data)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.

import logging
import os
import sqlite3
import unittest

from moat.mqtt.plugins.manager import BaseContext
from moat.mqtt.plugins.persistence import SQLitePlugin

formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
logging.basicConfig(level=logging.DEBUG, format=formatter)


class TestSQLitePlugin(unittest.TestCase):
    def test_create_tables(self):
        dbfile = os.path.join(os.path.dirname(os.path.realpath(__file__)), "test.db")
        context = BaseContext()
        context.logger = logging.getLogger(__name__)
        context.config = {"persistence": {"file": dbfile}}
        SQLitePlugin(context)

        conn = sqlite3.connect(dbfile)
        cursor = conn.cursor()
        rows = cursor.execute("SELECT name FROM sqlite_master where type = 'table'")
        tables = []
        for row in rows:
            tables.append(row[0])
        self.assertIn("session", tables)

    # def test_save_session(self):
    #     dbfile = os.path.join(os.path.dirname(os.path.realpath(__file__)), "test.db")
    #     context = BaseContext()
    #     context.logger = logging.getLogger(__name__)
    #     context.config = {
    #         'persistence': {
    #             'file': dbfile
    #         }
    #     }
    #     sql_plugin = SQLitePlugin(context)
    #     s = Session()
    #     s.client_id = 'test_save_session'
    #     ret = asyncio.run(sql_plugin.save_session(session=s))
    #
    #     conn = sqlite3.connect(dbfile)
    #     cursor = conn.cursor()
    #     row = cursor.execute("SELECT client_id FROM session where client_id = 'test_save_session'").fetchone()
    #     self.assertTrue(len(row) == 1)
    #     self.assertEqual(row[0], s.client_id)



# command line interface

import json
import logging
import os
import socket
import sys
from contextlib import AsyncExitStack

import anyio
import asyncclick as click
from asyncscope import main_scope
from moat.util import attrdict, combine_dict, yload

from .broker import create_broker
from .client import ConnectException, _codecs, open_mqttclient
from .utils import read_yaml_config
from .version import get_version

logger = logging.getLogger(__name__)


@click.group(short_help="MQTT client and broker")
async def cli():
    """
    Run MQTT commands
    """
    pass


@cli.command()
@click.pass_obj
async def broker(obj):
    """
    A basic MQTT broker that supports plug-ins and can use distkv as backend.
    """
    try:
        from moat.util import as_service
    except ImportError:
        as_service = None

    async with AsyncExitStack() as stack:
        await stack.enter_async_context(main_scope())
        await stack.enter_async_context(create_broker(obj.cfg.mqtt.broker))
        if as_service is not None:
            evt = await stack.enter_async_context(as_service())
            evt.set()
        while True:
            await anyio.sleep(99999)


def _gen_client_id():
    pid = os.getpid()
    hostname = socket.gethostname()
    return "moat_mqtt_pub/%d-%s" % (pid, hostname)


def _get_qos(args, cfg):
    res = args["qos"]
    if res is None:
        return cfg.default_qos


def _get_extra_headers(args, cfg):
    if args["extra_headers"]:
        hdrs = yload(args["extra_headers"])
        return combine_dict(hdrs, cfg.extra_headers)
    else:
        return cfg.extra_headers


def _get_message(args):
    codec = args["codec"] or "utf8"
    codec = _codecs[codec]()

    for m in args["msg"]:
        yield m.encode(encoding="utf-8")
    for m in args["msg_eval"]:
        yield codec.encode(eval(m))  # pylint: disable=eval-used
    if args["msg_lines"]:
        with open(args["msg_lines"], "r") as f:  # pylint: disable=unspecified-encoding
            for line in f:
                yield line.encode(encoding="utf-8")
    if args["msg_stdin_lines"]:
        for line in sys.stdin:
            if line:
                yield line.encode(encoding="utf-8")
    if args["msg_stdin"]:
        yield sys.stdin.buffer.read()
    if args["msg_stdin_eval"]:
        message = sys.stdin.read()
        yield codec.encode(eval(message))  # pylint: disable=eval-used


async def do_pub(client, args, cfg):
    logger.info("%s Connecting to broker", client.client_id)
    url = args["url"] or cfg.url

    await client.connect(
        uri=url,
        cleansession=args["clean_session"],
        cafile=args["ca_file"] or cfg.ca.file,
        capath=args["ca_path"] or cfg.ca.path,
        cadata=args["ca_data"] or cfg.ca.data,
        extra_headers=_get_extra_headers(args, cfg),
    )
    try:
        qos = _get_qos(args, cfg)
        topic = args["topic"]
        retain = args["retain"]
        if retain is None:
            retain = cfg.default_retain

        async with anyio.create_task_group() as tg:
            for message in _get_message(args):
                logger.info("%s Publishing to '%s'", client.client_id, topic)
                tg.start_soon(client.publish, topic, message, qos, retain)
        logger.info("%s Disconnected from broker", client.client_id)
    except KeyboardInterrupt:
        logger.info("%s Disconnected from broker", client.client_id)
    except ConnectException as ce:
        logger.fatal("connection to '%s' failed: %r", url, ce)
    finally:
        with anyio.fail_after(2, shield=True):
            await client.disconnect()


def fix_will(args, cfg):
    if args["will_topic"] and args["will_message"]:
        will = attrdict()
        will.topic = args["will_topic"]
        will.message = args["will_message"]
        will.qos = args["will_qos"]
        if will.qos is None:
            will.qos = cfg.will.qos
        will.retain = args["will_retain"]
        cfg.will = will
    if isinstance(cfg.will.message, str):
        cfg.will.message = cfg.will.message.encode("utf-8")


@cli.command()
@click.option("--url", help="Broker connection URL (musr conform to MQTT URI scheme")
@click.option("-i", "--client_id", help="string to use as client ID")
@click.option("-q", "--qos", type=click.IntRange(0, 2), help="Quality of service to use (0-2)")
@click.option("-r", "--retain", "retain", flag_value=True, help="Set the Retain flag")
@click.option("--no-retain", "retain", flag_value=False, help="Clear the Retain flag")
@click.option(
    "--default-retain",
    "retain",
    flag_value=None,
    help="Use the Retain flag's default",
    hidden=True,
)
@click.option("-t", "--topic", required=True, help="Message topic, '/'-separated")
@click.option("-m", "--msg", multiple=True, help="Message data (may be repeated)")
@click.option(
    "-M", "--msg-eval", multiple=True, help="Message data (Python, evaluated, may be repeated)"
)
@click.option(
    "-f", "--msg-lines", type=click.File("r"), help="File with messages (each line sent separately"
)
@click.option("-R", "--msg-stdin", is_flag=True, help="Single message from stdin")
@click.option(
    "-s", "--msg-stdin-lines", is_flag=True, help="Messages from stdin (each line sent separately"
)
@click.option(
    "-S",
    "--msg-stdin-eval",
    is_flag=True,
    help="Python code that evaluates to the message on stdin",
)
@click.option("-C", "--codec", help="Message codec (default UTF-8)")
@click.option("-k", "--keep-alive", type=float, help="Keep-alive timeout (seconds)")
@click.option("--clean-session", is_flag=True, help="Clean session on connect?")
@click.option("--ca-file", help="CA file")
@click.option("--ca-path", help="CA path")
@click.option("--ca-data", help="CA data")
@click.option("--will-topic", help="Topic to send to, when client exits")
@click.option("--will-message", help="Message to send, when client exits")
@click.option("--will-qos", type=int, help="QOS for Will message")
@click.option("--will-retain", is_flag=True, help="Retain Will message?")
@click.option("--extra-headers", type=click.File("r"), help="File to read extra MQTT headers from")
@click.pass_obj
async def pub(obj, **args):
    """Publish one or more MQTT messages"""
    if args["msg_stdin"] + args["msg_stdin_lines"] + args["msg_stdin_eval"] > 1:
        raise click.UsageError("You can only read from stdin once")
    cfg = obj.cfg.mqtt.client
    client_id = args["client_id"] or cfg.get("id", None) or _gen_client_id()

    if args["keep_alive"]:
        cfg["keep_alive"] = args["keep_alive"]

    fix_will(args, cfg)
    cfg.codec = "noop"

    async with open_mqttclient(client_id=client_id, config=cfg) as C:
        await do_pub(C, args, cfg)


async def do_sub(client, args, cfg):

    try:
        await client.connect(
            uri=args["url"] or cfg.url,
            cleansession=args["clean_session"],
            cafile=args["ca_file"] or cfg.ca.file,
            capath=args["ca_path"] or cfg.ca.path,
            cadata=args["ca_data"] or cfg.ca.data,
            extra_headers=_get_extra_headers(args, cfg),
        )
        async with anyio.create_task_group() as tg:
            for topic in args["topic"]:
                tg.start_soon(run_sub, client, topic, args, cfg)

    except KeyboardInterrupt:
        pass
    except ConnectException as ce:
        logger.fatal("connection to '%s' failed: %r", args["url"], ce)
    finally:
        with anyio.fail_after(2, shield=True):
            await client.disconnect()


async def run_sub(client, topic, args, cfg):
    qos = _get_qos(args, cfg)
    max_count = args["n_msg"]
    count = 0

    async with client.subscription(topic, qos) as subscr:
        async for message in subscr:
            count += 1
            print(message.topic, message.data, sep="\t")
            if max_count and count >= max_count:
                break


@cli.command()
@click.option("--url", help="Broker connection URL (musr conform to MQTT URI scheme")
@click.option("-i", "--client_id", help="string to use as client ID")
@click.option("-q", "--qos", type=click.IntRange(0, 2), help="Quality of service to use (0-2)")
@click.option(
    "-t",
    "--topic",
    multiple=True,
    help="Message topic, '/'-separated (can be used more than once)",
)
@click.option("-n", "--n_msg", type=int, default=0, help="Number of messages to read (per topic)")
@click.option("-C", "--codec", help="Message codec (default UTF-8)")
@click.option("-k", "--keep-alive", type=float, help="Keep-alive timeout (seconds)")
@click.option("--clean-session", is_flag=True, help="Clean session on connect?")
@click.option("--ca-file", help="CA file")
@click.option("--ca-path", help="CA path")
@click.option("--ca-data", help="CA data")
@click.option("--will-topic", help="Topic to send to, when client exits")
@click.option("--will-message", help="Message to send, when client exits")
@click.option("--will-qos", type=int, help="QOS for Will message")
@click.option("--will-retain", is_flag=True, help="Retain Will message?")
@click.option("--extra-headers", type=click.File("r"), help="File to read extra MQTT headers from")
@click.pass_obj
async def sub(obj, **args):
    """Subscribe to one or more MQTT topics"""
    cfg = obj.cfg.mqtt.client

    client_id = args["client_id"] or cfg.get("id", None) or _gen_client_id()

    if args["keep_alive"]:
        cfg["keep_alive"] = args["keep_alive"]

    fix_will(args, cfg)

    async with open_mqttclient(client_id=client_id, config=cfg, codec=args["codec"]) as C:
        await do_sub(C, args, cfg)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import unittest

from moat.mqtt.adapters import BufferAdapter
from moat.mqtt.mqtt.connect import ConnectPacket, ConnectPayload, ConnectVariableHeader
from moat.mqtt.mqtt.packet import CONNECT, MQTTFixedHeader

from .. import anyio_run


class ConnectPacketTest(unittest.TestCase):
    def test_decode_ok(self):
        data = b"\x10\x3e\x00\x04MQTT\x04\xce\x00\x00\x00\x0a0123456789\x00\x09WillTopic\x00\x0bWillMessage\x00\x04user\x00\x08password"
        stream = BufferAdapter(data)
        message = anyio_run(ConnectPacket.from_stream, stream)
        self.assertEqual(message.variable_header.proto_name, "MQTT")
        self.assertEqual(message.variable_header.proto_level, 4)
        self.assertTrue(message.variable_header.username_flag)
        self.assertTrue(message.variable_header.password_flag)
        self.assertFalse(message.variable_header.will_retain_flag)
        self.assertEqual(message.variable_header.will_qos, 1)
        self.assertTrue(message.variable_header.will_flag)
        self.assertTrue(message.variable_header.clean_session_flag)
        self.assertFalse(message.variable_header.reserved_flag)
        self.assertEqual(message.payload.client_id, "0123456789")
        self.assertEqual(message.payload.will_topic, "WillTopic")
        self.assertEqual(message.payload.will_message, b"WillMessage")
        self.assertEqual(message.payload.username, "user")
        self.assertEqual(message.payload.password, "password")

    def test_decode_ok_will_flag(self):
        data = (
            b"\x10\x26\x00\x04MQTT\x04\xca\x00\x00\x00\x0a0123456789\x00\x04user\x00\x08password"
        )
        stream = BufferAdapter(data)
        message = anyio_run(ConnectPacket.from_stream, stream)
        self.assertEqual(message.variable_header.proto_name, "MQTT")
        self.assertEqual(message.variable_header.proto_level, 4)
        self.assertTrue(message.variable_header.username_flag)
        self.assertTrue(message.variable_header.password_flag)
        self.assertFalse(message.variable_header.will_retain_flag)
        self.assertEqual(message.variable_header.will_qos, 1)
        self.assertFalse(message.variable_header.will_flag)
        self.assertTrue(message.variable_header.clean_session_flag)
        self.assertFalse(message.variable_header.reserved_flag)
        self.assertEqual(message.payload.client_id, "0123456789")
        self.assertEqual(message.payload.will_topic, None)
        self.assertEqual(message.payload.will_message, None)
        self.assertEqual(message.payload.username, "user")
        self.assertEqual(message.payload.password, "password")

    def test_decode_fail_reserved_flag(self):
        data = b"\x10\x3e\x00\x04MQTT\x04\xcf\x00\x00\x00\x0a0123456789\x00\x09WillTopic\x00\x0bWillMessage\x00\x04user\x00\x08password"
        stream = BufferAdapter(data)
        message = anyio_run(ConnectPacket.from_stream, stream)
        self.assertTrue(message.variable_header.reserved_flag)

    def test_decode_fail_miss_clientId(self):
        data = b"\x10\x0a\x00\x04MQTT\x04\xce\x00\x00"
        stream = BufferAdapter(data)
        message = anyio_run(ConnectPacket.from_stream, stream)
        self.assertIsNot(message.payload.client_id, None)

    def test_decode_fail_miss_willtopic(self):
        data = b"\x10\x16\x00\x04MQTT\x04\xce\x00\x00\x00\x0a0123456789"
        stream = BufferAdapter(data)
        message = anyio_run(ConnectPacket.from_stream, stream)
        self.assertIs(message.payload.will_topic, None)

    def test_decode_fail_miss_username(self):
        data = b"\x10\x2e\x00\x04MQTT\x04\xce\x00\x00\x00\x0a0123456789\x00\x09WillTopic\x00\x0bWillMessage"
        stream = BufferAdapter(data)
        message = anyio_run(ConnectPacket.from_stream, stream)
        self.assertIs(message.payload.username, None)

    def test_decode_fail_miss_password(self):
        data = b"\x10\x34\x00\x04MQTT\x04\xce\x00\x00\x00\x0a0123456789\x00\x09WillTopic\x00\x0bWillMessage\x00\x04user"
        stream = BufferAdapter(data)
        message = anyio_run(ConnectPacket.from_stream, stream)
        self.assertIs(message.payload.password, None)

    def test_encode(self):
        header = MQTTFixedHeader(CONNECT, 0x00, 0)
        variable_header = ConnectVariableHeader(0xCE, 0, "MQTT", 4)
        payload = ConnectPayload("0123456789", "WillTopic", b"WillMessage", "user", "password")
        message = ConnectPacket(header, variable_header, payload)
        encoded = message.to_bytes()
        self.assertEqual(
            encoded,
            b"\x10\x3e\x00\x04MQTT\x04\xce\x00\x00\x00\x0a0123456789\x00\x09WillTopic\x00\x0bWillMessage\x00\x04user\x00\x08password",
        )

    def test_getattr_ok(self):
        data = b"\x10\x3e\x00\x04MQTT\x04\xce\x00\x00\x00\x0a0123456789\x00\x09WillTopic\x00\x0bWillMessage\x00\x04user\x00\x08password"
        stream = BufferAdapter(data)
        message = anyio_run(ConnectPacket.from_stream, stream)
        self.assertEqual(message.variable_header.proto_name, "MQTT")
        self.assertEqual(message.proto_name, "MQTT")
        self.assertEqual(message.variable_header.proto_level, 4)
        self.assertEqual(message.proto_level, 4)
        self.assertTrue(message.variable_header.username_flag)
        self.assertTrue(message.username_flag)
        self.assertTrue(message.variable_header.password_flag)
        self.assertTrue(message.password_flag)
        self.assertFalse(message.variable_header.will_retain_flag)
        self.assertFalse(message.will_retain_flag)
        self.assertEqual(message.variable_header.will_qos, 1)
        self.assertEqual(message.will_qos, 1)
        self.assertTrue(message.variable_header.will_flag)
        self.assertTrue(message.will_flag)
        self.assertTrue(message.variable_header.clean_session_flag)
        self.assertTrue(message.clean_session_flag)
        self.assertFalse(message.variable_header.reserved_flag)
        self.assertFalse(message.reserved_flag)
        self.assertEqual(message.payload.client_id, "0123456789")
        self.assertEqual(message.client_id, "0123456789")
        self.assertEqual(message.payload.will_topic, "WillTopic")
        self.assertEqual(message.will_topic, "WillTopic")
        self.assertEqual(message.payload.will_message, b"WillMessage")
        self.assertEqual(message.will_message, b"WillMessage")
        self.assertEqual(message.payload.username, "user")
        self.assertEqual(message.username, "user")
        self.assertEqual(message.payload.password, "password")
        self.assertEqual(message.password, "password")



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
from ..errors import MoatMQTTException
from .connack import ConnackPacket
from .connect import ConnectPacket
from .disconnect import DisconnectPacket
from .packet import (
    CONNACK,
    CONNECT,
    DISCONNECT,
    PINGREQ,
    PINGRESP,
    PUBACK,
    PUBCOMP,
    PUBLISH,
    PUBREC,
    PUBREL,
    SUBACK,
    SUBSCRIBE,
    UNSUBACK,
    UNSUBSCRIBE,
    MQTTFixedHeader,
)
from .pingreq import PingReqPacket
from .pingresp import PingRespPacket
from .puback import PubackPacket
from .pubcomp import PubcompPacket
from .publish import PublishPacket
from .pubrec import PubrecPacket
from .pubrel import PubrelPacket
from .suback import SubackPacket
from .subscribe import SubscribePacket
from .unsuback import UnsubackPacket
from .unsubscribe import UnsubscribePacket

packet_dict = {
    CONNECT: ConnectPacket,
    CONNACK: ConnackPacket,
    PUBLISH: PublishPacket,
    PUBACK: PubackPacket,
    PUBREC: PubrecPacket,
    PUBREL: PubrelPacket,
    PUBCOMP: PubcompPacket,
    SUBSCRIBE: SubscribePacket,
    SUBACK: SubackPacket,
    UNSUBSCRIBE: UnsubscribePacket,
    UNSUBACK: UnsubackPacket,
    PINGREQ: PingReqPacket,
    PINGRESP: PingRespPacket,
    DISCONNECT: DisconnectPacket,
}


def packet_class(fixed_header: MQTTFixedHeader):
    try:
        cls = packet_dict[fixed_header.packet_type]
        return cls
    except KeyError:
        raise MoatMQTTException(  # pylint:disable=W0707
            "Unexpected packet Type '%s'" % fixed_header.packet_type
        )



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import logging

import anyio
import attr
import yaml

from .errors import InvalidStateError

logger = logging.getLogger(__name__)


def format_client_message(session=None, address=None, port=None):
    if session:
        return "(client id=%s)" % session.client_id
    elif address is not None and port is not None:
        return "(client @=%s:%d)" % (address, port)
    else:
        return "(unknown client)"


def gen_client_id():
    """
    Generates random client ID
    :return:
    """
    import random

    gen_id = "moat-mqtt-"

    for _ in range(16):
        gen_id += chr(random.randint(0, 26) + 97)
    return gen_id


def read_yaml_config(config_file):
    config = None
    try:
        with open(config_file, "r") as stream:  # pylint: disable=unspecified-encoding
            config = yaml.safe_load(stream)
    except yaml.YAMLError as exc:
        logger.error("Invalid config_file %s: %r", config_file, exc)
    return config


# utility code


class CancelledError(RuntimeError):
    # This intentionally does not descend from any toolkit's cancellation exception
    # (much less from all of them)
    pass


@attr.s
class Future:
    """A waitable value useful for inter-task synchronization,
    inspired by :class:`threading.Event` and :class:`asyncio.Future`.

    An event object manages an internal value, which is initially
    unset, and a task can wait for it to become True.

    Note that the value can only be read once.
    """

    event = attr.ib(factory=anyio.Event, init=False)
    value = attr.ib(default=None, init=False)

    async def set(self, value):
        """Set the result to return this value, and wake any waiting task."""
        if self.event.is_set():
            raise InvalidStateError("Value already set")
        self.value = value
        self.event.set()

    async def set_error(self, exc):
        """Set the result to raise this exceptio, and wake any waiting task."""
        if self.event.is_set():
            raise InvalidStateError("Value already set")
        self.value = exc
        self.event.set()

    def is_set(self):
        """Check whether the event has occurred."""
        return self.value is not None

    async def cancel(self):
        """Send a cancelation to the recipient."""
        await self.set_error(CancelledError())

    async def get(self):
        """Block until the value is set.

        If it's already set, then this method returns immediately.

        The value can only be read once.
        """
        await self.event.wait()
        if isinstance(self.value, BaseException):
            raise self.value
        return self.value

    def done(self):
        """Report whether this Future has been set."""
        return self.event.is_set()


def match_topic(topic, subscription):
    """
    Match @topic to @subscription. Both must be lists/tuples.
    """
    if isinstance(topic, str) or isinstance(subscription, str):
        raise RuntimeError("Subscriptions need to be pre-split")
    if topic[0].startswith("$") != subscription[0].startswith("$"):
        return False
    if len(topic) < len(subscription):
        return False
    if len(topic) > len(subscription) and subscription[-1] != "#":
        return False
    for a, b in zip(topic, subscription):
        if a != b and b not in ("+", "#"):
            return False
    return True



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import unittest

from moat.mqtt.adapters import BufferAdapter
from moat.mqtt.codecs import bytes_to_hex_str, bytes_to_int, decode_string, encode_string

from . import anyio_run


class TestCodecs(unittest.TestCase):
    def test_bytes_to_hex_str(self):
        ret = bytes_to_hex_str(b"\x7f")
        self.assertEqual(ret, "0x7f")

    def test_bytes_to_int(self):
        ret = bytes_to_int(b"\x7f")
        self.assertEqual(ret, 127)
        ret = bytes_to_int(b"\xff\xff")
        self.assertEqual(ret, 65535)

    def test_decode_string(self):
        async def test_coro():
            stream = BufferAdapter(b"\x00\x02AA")
            ret = await decode_string(stream)
            self.assertEqual(ret, "AA")

        anyio_run(test_coro)

    def test_encode_string(self):
        encoded = encode_string("AA")
        self.assertEqual(b"\x00\x02AA", encoded)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import anyio

from ..codecs import bytes_to_int, decode_string, encode_string, int_to_bytes, read_or_raise
from ..errors import MoatMQTTException, NoDataException
from .packet import (
    SUBSCRIBE,
    MQTTFixedHeader,
    MQTTPacket,
    MQTTPayload,
    MQTTVariableHeader,
    PacketIdVariableHeader,
)


class SubscribePayload(MQTTPayload):

    __slots__ = ("topics",)

    def __init__(self, topics=()):
        super().__init__()
        self.topics = topics

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        out = b""
        for topic in self.topics:
            out += encode_string(topic[0])
            out += int_to_bytes(topic[1], 1)
        return out

    @classmethod
    async def from_stream(
        cls,
        reader: anyio.abc.ByteStream,
        fixed_header: MQTTFixedHeader,
        variable_header: MQTTVariableHeader,
    ):
        topics = []
        payload_length = fixed_header.remaining_length - variable_header.bytes_length
        read_bytes = 0
        while read_bytes < payload_length:
            try:
                topic = await decode_string(reader)
                qos_byte = await read_or_raise(reader, 1)
                qos = bytes_to_int(qos_byte)
                topics.append((topic, qos))
                read_bytes += 2 + len(topic.encode("utf-8")) + 1
            except NoDataException:
                break
        return cls(topics)

    def __repr__(self):
        return type(self).__name__ + "(topics={0!r})".format(self.topics)


class SubscribePacket(MQTTPacket):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = SubscribePayload

    def __init__(
        self,
        fixed: MQTTFixedHeader = None,
        variable_header: PacketIdVariableHeader = None,
        payload=None,
    ):
        if fixed is None:
            header = MQTTFixedHeader(SUBSCRIBE, 0x02)  # [MQTT-3.8.1-1]
        else:
            if fixed.packet_type != SUBSCRIBE:
                raise MoatMQTTException(
                    "Invalid fixed packet type %s for SubscribePacket init" % fixed.packet_type
                )
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @classmethod
    def build(cls, topics, packet_id):
        v_header = PacketIdVariableHeader(packet_id)
        payload = SubscribePayload(topics)
        return SubscribePacket(variable_header=v_header, payload=payload)



"""
This module contains code that helps with DistKV testing.
"""
import os
from contextlib import asynccontextmanager
from functools import partial

import anyio
from distkv.client import open_client
from distkv.server import Server as _Server

from .broker import create_broker


class Server(_Server):
    @asynccontextmanager
    async def test_client(self):
        """
        An async context manager that returns a client that's connected to
        this server.
        """
        async with open_client(connect=dict(host="127.0.0.1", port=self.distkv_port)) as client:
            yield client


@asynccontextmanager
async def test_server(mqtt_port: int = None, distkv_port: int = None):
    """
    An async context manager which creates a stand-alone DistKV server.

    The server has a `test_client` method: an async context manager that
    returns a client taht's connected to this server.

    Ports are allocated based on the current process's PID.
    """
    if mqtt_port is None:
        mqtt_port = 40000 + os.getpid() % 10000
    if distkv_port is None:
        distkv_port = 40000 + (os.getpid() + 1) % 10000

    broker_cfg = {
        "listeners": {"default": {"type": "tcp", "bind": "127.0.0.1:%d" % mqtt_port}},
        "timeout-disconnect-delay": 2,
        "auth": {"allow-anonymous": True, "password-file": None},
    }
    server_cfg = {
        "server": {
            "bind_default": {"host": "127.0.0.1", "port": distkv_port},
            "backend": "mqtt",
            "mqtt": {"uri": "mqtt://127.0.0.1:%d/" % mqtt_port},
        }
    }

    server = Server(name="gpio_test", cfg=server_cfg, init="GPIO")
    async with create_broker(config=broker_cfg) as broker:
        evt = anyio.Event()
        broker._tg.start_soon(partial(server.serve, ready_evt=evt))
        await evt.wait()

        server.distkv_port = distkv_port  # pylint: disable=attribute-defined-outside-init
        yield server


@asynccontextmanager
async def test_client(mqtt_port: int = None, distkv_port: int = None):
    """
    An async context manager which creates a stand-alone DistKV client.
    """
    async with test_server(mqtt_port=mqtt_port, distkv_port=distkv_port) as s:
        async with s.test_client() as c:
            yield c



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import logging
import os
import unittest
from unittest.mock import MagicMock, call, patch

import anyio
import pytest

from moat.mqtt.adapters import StreamAdapter
from moat.mqtt.broker import (
    EVENT_BROKER_CLIENT_CONNECTED,
    EVENT_BROKER_CLIENT_DISCONNECTED,
    EVENT_BROKER_CLIENT_SUBSCRIBED,
    EVENT_BROKER_CLIENT_UNSUBSCRIBED,
    EVENT_BROKER_MESSAGE_RECEIVED,
    EVENT_BROKER_POST_SHUTDOWN,
    EVENT_BROKER_POST_START,
    EVENT_BROKER_PRE_SHUTDOWN,
    EVENT_BROKER_PRE_START,
    create_broker,
)
from moat.mqtt.client import ConnectException, open_mqttclient
from moat.mqtt.mqtt import (
    ConnackPacket,
    ConnectPacket,
    DisconnectPacket,
    PublishPacket,
    PubrecPacket,
    PubrelPacket,
)
from moat.mqtt.mqtt.connect import ConnectPayload, ConnectVariableHeader
from moat.mqtt.mqtt.constants import QOS_0, QOS_1, QOS_2

from . import anyio_run

formatter = "%(asctime)s %(name)s:%(lineno)d %(levelname)s - %(message)s"
logging.basicConfig(level=logging.INFO, format=formatter)
log = logging.getLogger(__name__)

PORT = 40000 + (os.getpid() + 3) % 10000
URL = "mqtt://127.0.0.1:%d/" % PORT

test_config = {
    "listeners": {
        "default": {"type": "tcp", "bind": "127.0.0.1:%d" % PORT, "max_connections": 10}
    },
    "sys_interval": 0,
    "auth": {"allow-anonymous": True},
}


class AsyncMock(MagicMock):
    def __await__(self):
        async def foo():
            return self

        return foo().__await__()


class BrokerTest(unittest.TestCase):
    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_start_stop(self, MockPluginManager):  # pylint: disable=unused-argument
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                self.assertDictEqual(broker._sessions, {})
                self.assertIn("default", broker._servers)
                MockPluginManager.assert_has_calls(
                    [
                        call().fire_event(EVENT_BROKER_PRE_START),
                        call().fire_event(EVENT_BROKER_POST_START),
                    ],
                    any_order=True,
                )
                MockPluginManager.reset_mock()
            MockPluginManager.assert_has_calls(
                [
                    call().fire_event(EVENT_BROKER_PRE_SHUTDOWN),
                    call().fire_event(EVENT_BROKER_POST_SHUTDOWN),
                ],
                any_order=True,
            )
            self.assertTrue(broker.transitions.is_stopped())

        anyio_run(test_coro, backend="trio")

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_connect(self, MockPluginManager):
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient() as client:
                    ret = await client.connect(URL)
                    self.assertEqual(ret, 0)
                    self.assertEqual(len(broker._sessions), 1)
                    self.assertIn(client.session.client_id, broker._sessions)
                await anyio.sleep(0.1)  # let the broker task process the packet
            self.assertTrue(broker.transitions.is_stopped())
            self.assertDictEqual(broker._sessions, {})
            MockPluginManager.assert_has_calls(
                [
                    call().fire_event(
                        EVENT_BROKER_CLIENT_CONNECTED,
                        client_id=client.session.client_id,
                    ),
                    call().fire_event(
                        EVENT_BROKER_CLIENT_DISCONNECTED,
                        client_id=client.session.client_id,
                    ),
                ],
                any_order=True,
            )

        anyio_run(test_coro)

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_connect_will_flag(self, MockPluginManager):  # pylint: disable=unused-argument
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())

                async with await anyio.connect_tcp("127.0.0.1", PORT) as conn:
                    stream = StreamAdapter(conn)

                    vh = ConnectVariableHeader()
                    payload = ConnectPayload()

                    vh.keep_alive = 10
                    vh.clean_session_flag = False
                    vh.will_retain_flag = False
                    vh.will_flag = True
                    vh.will_qos = QOS_0
                    payload.client_id = "test_id"
                    payload.will_message = b"test"
                    payload.will_topic = "/topic"
                    connect = ConnectPacket(vh=vh, payload=payload)
                    await connect.to_stream(stream)
                    await ConnackPacket.from_stream(stream)

                    disconnect = DisconnectPacket()
                    await disconnect.to_stream(stream)

            self.assertTrue(broker.transitions.is_stopped())
            self.assertDictEqual(broker._sessions, {})

        anyio_run(test_coro, backend="trio")

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_connect_clean_session_false(
        self, MockPluginManager
    ):  # pylint: disable=unused-argument
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient(
                    client_id="", config={"auto_reconnect": False}
                ) as client:
                    return_code = None
                    with pytest.raises(ConnectException) as ce:
                        await client.connect(URL, cleansession=False)
                    return_code = ce.value.return_code
                    self.assertEqual(return_code, 0x02)
                    self.assertNotIn(client.session.client_id, broker._sessions)

        anyio_run(test_coro)

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_subscribe(self, MockPluginManager):
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient() as client:
                    ret = await client.connect(URL)
                    self.assertEqual(ret, 0)
                    await client.subscribe([("/topic", QOS_0)])

                    # Test if the client test client subscription is registered
                    subs = broker._subscriptions[("", "topic")]
                    self.assertEqual(len(subs), 1)
                    (s, qos) = subs[0]
                    self.assertEqual(s, client.session)
                    self.assertEqual(qos, QOS_0)

            self.assertTrue(broker.transitions.is_stopped())
            MockPluginManager.assert_has_calls(
                [
                    call().fire_event(
                        EVENT_BROKER_CLIENT_SUBSCRIBED,
                        client_id=client.session.client_id,
                        topic="/topic",
                        qos=QOS_0,
                    )
                ],
                any_order=True,
            )

        anyio_run(test_coro, backend="trio")

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_subscribe_twice(self, MockPluginManager):
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient() as client:
                    ret = await client.connect(URL)
                    self.assertEqual(ret, 0)
                    await client.subscribe([("/topic", QOS_0)])

                    # Test if the client test client subscription is registered
                    subs = broker._subscriptions[("", "topic")]
                    self.assertEqual(len(subs), 1)
                    (s, qos) = subs[0]
                    self.assertEqual(s, client.session)
                    self.assertEqual(qos, QOS_0)

                    await client.subscribe([("/topic", QOS_0)])
                    self.assertEqual(len(subs), 1)
                    (s, qos) = subs[0]
                    self.assertEqual(s, client.session)
                    self.assertEqual(qos, QOS_0)

            self.assertTrue(broker.transitions.is_stopped())
            MockPluginManager.assert_has_calls(
                [
                    call().fire_event(
                        EVENT_BROKER_CLIENT_SUBSCRIBED,
                        client_id=client.session.client_id,
                        topic="/topic",
                        qos=QOS_0,
                    )
                ],
                any_order=True,
            )

        anyio_run(test_coro, backend="trio")

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_unsubscribe(self, MockPluginManager):
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient() as client:
                    ret = await client.connect(URL)
                    self.assertEqual(ret, 0)
                    await client.subscribe([("/topic", QOS_0)])

                    # Test if the client test client subscription is registered
                    subs = broker._subscriptions[("", "topic")]
                    self.assertEqual(len(subs), 1)
                    (s, qos) = subs[0]
                    self.assertEqual(s, client.session)
                    self.assertEqual(qos, QOS_0)

                    await client.unsubscribe(["/topic"])
                    self.assertEqual(broker._subscriptions[("", "topic")], [])
            self.assertTrue(broker.transitions.is_stopped())
            MockPluginManager.assert_has_calls(
                [
                    call().fire_event(
                        EVENT_BROKER_CLIENT_SUBSCRIBED,
                        client_id=client.session.client_id,
                        topic="/topic",
                        qos=QOS_0,
                    ),
                    call().fire_event(
                        EVENT_BROKER_CLIENT_UNSUBSCRIBED,
                        client_id=client.session.client_id,
                        topic="/topic",
                    ),
                ],
                any_order=True,
            )

        anyio_run(test_coro, backend="trio")

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_publish(self, MockPluginManager):
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient() as pub_client:
                    ret = await pub_client.connect(URL)
                    self.assertEqual(ret, 0)

                    ret_message = await pub_client.publish("/topic", b"data", QOS_0)
                await anyio.sleep(0.1)  # let the broker task process the packet
                self.assertEqual(broker._retained_messages, {})

            self.assertTrue(broker.transitions.is_stopped())
            MockPluginManager.assert_has_calls(
                [
                    call().fire_event(
                        EVENT_BROKER_MESSAGE_RECEIVED,
                        client_id=pub_client.session.client_id,
                        message=ret_message,
                    )
                ],
                any_order=True,
            )

        anyio_run(test_coro)

    # @patch('moat.mqtt.broker.PluginManager', new_callable=AsyncMock)
    def test_client_publish_dup(self):
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())

                async with await anyio.connect_tcp("127.0.0.1", PORT) as conn:
                    stream = StreamAdapter(conn)

                    vh = ConnectVariableHeader()
                    payload = ConnectPayload()

                    vh.keep_alive = 10
                    vh.clean_session_flag = False
                    vh.will_retain_flag = False
                    payload.client_id = "test_id"
                    connect = ConnectPacket(vh=vh, payload=payload)
                    await connect.to_stream(stream)
                    await ConnackPacket.from_stream(stream)

                    publish_1 = PublishPacket.build("/test", b"data", 1, False, QOS_2, False)
                    await publish_1.to_stream(stream)
                    await PubrecPacket.from_stream(stream)

                    publish_dup = PublishPacket.build("/test", b"data", 1, True, QOS_2, False)
                    await publish_dup.to_stream(stream)
                    await PubrecPacket.from_stream(stream)
                    pubrel = PubrelPacket.build(1)
                    await pubrel.to_stream(stream)
                    # await PubcompPacket.from_stream(stream)

                    disconnect = DisconnectPacket()
                    await disconnect.to_stream(stream)

        anyio_run(test_coro, backend="trio")

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_publish_invalid_topic(
        self, MockPluginManager
    ):  # pylint: disable=unused-argument
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient() as pub_client:
                    ret = await pub_client.connect(URL)
                    self.assertEqual(ret, 0)

                    await pub_client.publish("/+", b"data", QOS_0)

            self.assertTrue(broker.transitions.is_stopped())

        anyio_run(test_coro)

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_publish_big(self, MockPluginManager):
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient() as pub_client:
                    ret = await pub_client.connect(URL)
                    self.assertEqual(ret, 0)

                    ret_message = await pub_client.publish(
                        "/topic", bytearray(b"\x99" * 256 * 1024), QOS_2
                    )
                self.assertEqual(broker._retained_messages, {})

            self.assertTrue(broker.transitions.is_stopped())
            MockPluginManager.assert_has_calls(
                [
                    call().fire_event(
                        EVENT_BROKER_MESSAGE_RECEIVED,
                        client_id=pub_client.session.client_id,
                        message=ret_message,
                    )
                ],
                any_order=True,
            )

        anyio_run(test_coro, backend="trio")

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_publish_retain(self, MockPluginManager):  # pylint: disable=unused-argument
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())

                async with open_mqttclient() as pub_client:
                    ret = await pub_client.connect(URL)
                    self.assertEqual(ret, 0)
                    await pub_client.publish("/topic", b"data", QOS_0, retain=True)
                await anyio.sleep(0.1)  # let the broker task process the packet
                self.assertIn("/topic", broker._retained_messages)
                retained_message = broker._retained_messages["/topic"]
                self.assertEqual(retained_message.source_session, pub_client.session)
                self.assertEqual(retained_message.topic, "/topic")
                self.assertEqual(retained_message.data, b"data")
                self.assertEqual(retained_message.qos, QOS_0)
            self.assertTrue(broker.transitions.is_stopped())

        anyio_run(test_coro)

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_publish_retain_delete(
        self, MockPluginManager
    ):  # pylint: disable=unused-argument
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())

                async with open_mqttclient() as pub_client:
                    ret = await pub_client.connect(URL)
                    self.assertEqual(ret, 0)
                    await pub_client.publish("/topic", b"", QOS_0, retain=True)
                await anyio.sleep(0.1)  # let the broker task process the packet
                self.assertNotIn("/topic", broker._retained_messages)
            self.assertTrue(broker.transitions.is_stopped())

        anyio_run(test_coro)

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_subscribe_publish(self, MockPluginManager):  # pylint: disable=unused-argument
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient() as sub_client:
                    await sub_client.connect(URL)
                    ret = await sub_client.subscribe(
                        [("/qos0", QOS_0), ("/qos1", QOS_1), ("/qos2", QOS_2)]
                    )
                    self.assertEqual(ret, [QOS_0, QOS_1, QOS_2])

                    await self._client_publish("/qos0", b"data", QOS_0)
                    await self._client_publish("/qos1", b"data", QOS_1)
                    await self._client_publish("/qos2", b"data", QOS_2)
                    for qos in [QOS_0, QOS_1, QOS_2]:
                        message = await sub_client.deliver_message()
                        self.assertIsNotNone(message)
                        self.assertEqual(message.topic, "/qos%s" % qos)
                        self.assertEqual(message.data, b"data")
                        self.assertEqual(message.qos, qos)
            self.assertTrue(broker.transitions.is_stopped())

        anyio_run(test_coro)

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_subscribe_invalid(self, MockPluginManager):  # pylint: disable=unused-argument
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient() as sub_client:
                    await sub_client.connect(URL)
                    ret = await sub_client.subscribe(
                        [
                            ("+", QOS_0),
                            ("+/tennis/#", QOS_0),
                            ("sport+", QOS_0),
                            ("sport/+/player1", QOS_0),
                        ]
                    )
                    self.assertEqual(ret, [QOS_0, QOS_0, 0x80, QOS_0])

            self.assertTrue(broker.transitions.is_stopped())

        anyio_run(test_coro, backend="trio")

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_subscribe_publish_dollar_topic_1(
        self, MockPluginManager
    ):  # pylint: disable=unused-argument
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient() as sub_client:
                    await sub_client.connect(URL)
                    ret = await sub_client.subscribe([("#", QOS_0)])
                    self.assertEqual(ret, [QOS_0])

                    await self._client_publish("/topic", b"data", QOS_0)
                    message = await sub_client.deliver_message()
                    self.assertIsNotNone(message)

                    await self._client_publish("$topic", b"data", QOS_0)
                    message = None
                    with self.assertRaises(TimeoutError):
                        with anyio.fail_after(1):
                            message = await sub_client.deliver_message()
                    self.assertIsNone(message)
            self.assertTrue(broker.transitions.is_stopped())

        anyio_run(test_coro)

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_subscribe_publish_dollar_topic_2(
        self, MockPluginManager
    ):  # pylint: disable=unused-argument
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient() as sub_client:
                    await sub_client.connect(URL)
                    ret = await sub_client.subscribe([("+/monitor/Clients", QOS_0)])
                    self.assertEqual(ret, [QOS_0])

                    await self._client_publish("test/monitor/Clients", b"data", QOS_0)
                    message = await sub_client.deliver_message()
                    self.assertIsNotNone(message)

                    await self._client_publish("$SYS/monitor/Clients", b"data", QOS_0)
                    message = None
                    with self.assertRaises(TimeoutError):
                        with anyio.fail_after(1):
                            message = await sub_client.deliver_message()
                    self.assertIsNone(message)
            self.assertTrue(broker.transitions.is_stopped())

        anyio_run(test_coro)

    @patch("moat.mqtt.broker.PluginManager", new_callable=AsyncMock)
    def test_client_publish_retain_subscribe(
        self, MockPluginManager
    ):  # pylint: disable=unused-argument
        async def test_coro():
            async with create_broker(
                test_config, plugin_namespace="moat.mqtt.test.plugins"
            ) as broker:
                broker.plugins_manager._tg = broker._tg
                self.assertTrue(broker.transitions.is_started())
                async with open_mqttclient() as sub_client:
                    await sub_client.connect(URL, cleansession=False)
                    ret = await sub_client.subscribe(
                        [("/qos0", QOS_0), ("/qos1", QOS_1), ("/qos2", QOS_2)]
                    )
                    self.assertEqual(ret, [QOS_0, QOS_1, QOS_2])
                    await sub_client.disconnect()

                    await self._client_publish("/qos0", b"data", QOS_0, retain=True)
                    await self._client_publish("/qos1", b"data", QOS_1, retain=True)
                    await self._client_publish("/qos2", b"data", QOS_2, retain=True)
                    await sub_client.reconnect()
                    for qos in [QOS_0, QOS_1, QOS_2]:
                        log.debug("TEST QOS: %d", qos)
                        message = await sub_client.deliver_message()
                        log.debug("Message: %r", message.publish_packet)
                        self.assertIsNotNone(message)
                        self.assertEqual(message.topic, "/qos%s" % qos)
                        self.assertEqual(message.data, b"data")
                        self.assertEqual(message.qos, qos)
            self.assertTrue(broker.transitions.is_stopped())

        anyio_run(test_coro)

    async def _client_publish(self, topic, data, qos, retain=False):
        async with open_mqttclient() as pub_client:
            ret = await pub_client.connect(URL)
            self.assertEqual(ret, 0)
            ret = await pub_client.publish(topic, data, qos, retain)
        return ret



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
from struct import pack, unpack

import anyio

from .errors import NoDataException

try:
    IncompleteRead = anyio.exceptions.IncompleteRead
except AttributeError:
    IncompleteRead = anyio.IncompleteRead

# Collection of basic codecs for the messages' payload

import msgpack
import simplejson as json


def bytes_to_hex_str(data):
    """
    converts a sequence of bytes into its displayable hex representation, ie: 0x??????
    :param data: byte sequence
    :return: Hexadecimal displayable representation
    """
    return "0x" + "".join(format(b, "02x") for b in data)


def bytes_to_int(data):
    """
    convert a sequence of bytes to an integer using big endian byte ordering
    :param data: byte sequence
    :return: integer value
    """
    return int.from_bytes(data, byteorder="big")


def int_to_bytes(int_value: int, length: int) -> bytes:
    """
    convert an integer to a sequence of bytes using big endian byte ordering
    :param int_value: integer value to convert
    :param length: (optional) byte length
    :return: byte sequence
    """
    if length == 1:
        fmt = "!B"
    elif length == 2:
        fmt = "!H"
    return pack(fmt, int_value)


async def read_or_raise(reader, n=-1):
    """
    Read a given byte number from Stream. NoDataException is raised if read gives no data
    :param reader: reader adapter
    :param n: number of bytes to read
    :return: bytes read
    """
    try:
        data = await reader.read(n)
    except (IncompleteRead, ConnectionResetError):
        data = None
    if not data:
        raise NoDataException("No more data")
    return data


async def decode_string(reader) -> bytes:
    """
    Read a string from a reader and decode it according to MQTT string specification
    :param reader: Stream reader
    :return: UTF-8 string read from stream
    """
    return (await decode_data_with_length(reader)).decode("utf-8")


async def decode_data_with_length(reader) -> bytes:
    """
    Read data from a reader. Data is prefixed with 2 bytes length
    :param reader: Stream reader
    :return: bytes read from stream (without length)
    """
    length_bytes = await read_or_raise(reader, 2)
    bytes_length = unpack("!H", length_bytes)
    data = await read_or_raise(reader, bytes_length[0])
    return data


def encode_string(string: str) -> bytes:
    return encode_data_with_length(string.encode("utf-8"))


def encode_data_with_length(data: bytes) -> bytes:
    data_length = len(data)
    return int_to_bytes(data_length, 2) + data


async def decode_packet_id(reader) -> int:
    """
    Read a packet ID as 2-bytes int from stream according to MQTT specification (2.3.1)
    :param reader: Stream reader
    :return: Packet ID
    """
    packet_id_bytes = await read_or_raise(reader, 2)
    packet_id = unpack("!H", packet_id_bytes)
    return packet_id[0]


def int_to_bytes_str(value: int) -> bytes:
    """
    Converts a int value to a bytes array containing the numeric character.
    Ex: 123 -> b'123'
    :param value: int value to convert
    :return: bytes array
    """
    return str(value).encode("utf-8")


class BaseCodec:
    name = None

    def __init__(self, name=None):
        if name is not None and name != self.name:
            raise RuntimeError("Codec name mismatch")

    def encode(self, data):
        raise RuntimeError("You need to override me")

    def decode(self, data):
        raise RuntimeError("You need to override me")


# pylint doesn't (want to) understand "staticmethod"
# https://github.com/PyCQA/pylint/issues/6019
# pylint: disable=arguments-differ


class NoopCodec(BaseCodec):
    """A codec that does nothing.

    Your payload needs to consist of bytes.
    """

    name = "noop"

    @staticmethod
    def encode(data):
        assert isinstance(data, (bytearray, bytes))
        return data

    @staticmethod
    def decode(data):
        return data


class UTF8Codec(BaseCodec):
    """A codec that translates to UTF-8 strings.

    Your payload needs to be a single string.

    This codec will *not* stringify other data types for you.
    """

    name = "utf8"

    @staticmethod
    def encode(data):
        return data.encode("utf-8")

    @staticmethod
    def decode(data):
        return data.decode("utf-8")


class MsgPackCodec(BaseCodec):
    """A codec that encodes to "msgpack"-encoded bytestring.

    Your payload must consist of whatever "msgpack" accepts.

    Args:
      ``use_bin_type``: Bytestrings are encoded as byte arrays.
                        Defaults to ``True``. If ``False``,
                        bytes are transmitted as string types and input
                        strings are not UTF8-decoded. Use this if your
                        network contains clients which use
                        non-UTF8-encoded strings (this violates the
                        MsgPack specification).
      ``use_list``: if ``True``, lists and tuples are returned as lists
                    (i.e. modifyable). Defaults to ``False``, which uses
                    immutable tuples (this is faster).
    """

    name = "msgpack"

    use_bin_type = True
    use_list = False

    def __init__(self, use_bin_type=None, use_list=None, **kw):
        super().__init__(**kw)
        if use_bin_type is not None:
            self.use_bin_type = use_bin_type
        if use_list is not None:
            self.use_list = use_list

    def encode(self, data):
        return msgpack.packb(data, use_bin_type=self.use_bin_type)

    def decode(self, data):
        return msgpack.unpackb(data, raw=not self.use_bin_type, use_list=self.use_list)
        # raw=False would try to decode input bytes, which is not what you
        # want when the input is a bytestring. So we only use that if
        # bytestring input has been marked as such.


class JSONCodec(BaseCodec):
    """A codec that encodes to JSON+utf8.

    Your payload must consist of whatever simplejson accepts.
    """

    name = "json"

    @staticmethod
    def encode(data):
        return json.dumps(data).encode("utf-8")

    @staticmethod
    def decode(data):
        return json.loads(data.decode("utf-8"))


class MsgPackJSONCodec(BaseCodec):
    """A codec that encodes JSON strings to MsgPack.

    Your payload must consist of a valid JSON string. The wire format will
    be msgpack.

    This codec is useful for legacy code which wants messages as JSON strings
    when the rest of your world talks msgpack.
    """

    name = "msgpack+json"

    use_bin_type = True
    use_list = False

    def __init__(self, use_bin_type=None, use_list=None, **kw):
        super().__init__(**kw)
        if use_bin_type is not None:
            self.use_bin_type = use_bin_type
        if use_list is not None:
            self.use_list = use_list

    def encode(self, data):
        return msgpack.packb(
            json.loads(data), use_bin_type=self.use_bin_type, use_list=self.use_list
        )

    def decode(self, data):
        return json.dumps(msgpack.unpackb(data, raw=not self.use_bin_type, use_list=self.use_list))


class BoolCodec(BaseCodec):
    """A codec that recognizes two strings ("on" and "off") and bool-ifies them.

    Or maybe three ("null") but you need to tell it to do that.
    """

    name = "bool"

    on = "on"
    off = "off"
    null = None

    def __init__(self, on=None, off=None, null=None, **kw):
        super().__init__(**kw)
        if on is not None:
            self.on = on.encode("utf-8")
        if off is not None:
            self.off = off.encode("utf-8")
        if null is not None:
            self.null = null.encode("utf-8")

    def encode(self, data):
        if data is None and self.null is not None:
            return self.null
        if data == 0:
            return self.off
        if data == 1:
            return self.on
        raise ValueError(data)

    def decode(self, data):
        if data == self.on:
            return True
        if data == self.off:
            return False
        if self.null is not None and data == self.null:
            return None
        raise ValueError(data)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import anyio
from moat.util import create_queue

from ...plugins.manager import PluginManager
from ...session import Session
from ...utils import Future
from ..connack import ConnackPacket
from ..connect import ConnectPacket, ConnectPayload, ConnectVariableHeader
from ..disconnect import DisconnectPacket
from ..pingreq import PingReqPacket
from ..pingresp import PingRespPacket
from ..suback import SubackPacket
from ..subscribe import SubscribePacket
from ..unsuback import UnsubackPacket
from ..unsubscribe import UnsubscribePacket
from .handler import EVENT_MQTT_PACKET_RECEIVED, ProtocolHandler


class ClientProtocolHandler(ProtocolHandler):
    def __init__(self, plugins_manager: PluginManager, session: Session = None):
        super().__init__(plugins_manager, session)
        self._ping_task = None
        self._pingresp_queue = create_queue(9999)
        self._subscriptions_waiter = dict()
        self._unsubscriptions_waiter = dict()
        self._disconnect_waiter = None

    async def stop(self):
        try:
            await super().stop()
        finally:
            with anyio.fail_after(2, shield=True):
                t, self._ping_task = self._ping_task, None
                if t:
                    self.logger.debug("Cancel ping task")
                    await t.cancel()

    def _build_connect_packet(self):
        vh = ConnectVariableHeader()
        payload = ConnectPayload()

        vh.keep_alive = self.session.keep_alive
        vh.clean_session_flag = self.session.clean_session
        vh.will_retain_flag = self.session.will_retain
        payload.client_id = self.session.client_id

        if self.session.username:
            vh.username_flag = True
            payload.username = self.session.username
        else:
            vh.username_flag = False

        if self.session.password:
            vh.password_flag = True
            payload.password = self.session.password
        else:
            vh.password_flag = False
        if self.session.will_flag:
            vh.will_flag = True
            vh.will_qos = self.session.will_qos
            payload.will_message = self.session.will_message
            payload.will_topic = self.session.will_topic
        else:
            vh.will_flag = False

        packet = ConnectPacket(vh=vh, payload=payload)
        return packet

    async def mqtt_connect(self):
        connect_packet = self._build_connect_packet()
        await self._send_packet(connect_packet)
        connack = await ConnackPacket.from_stream(self.stream)
        self.logger.debug("< C %r", connack)
        await self.plugins_manager.fire_event(
            EVENT_MQTT_PACKET_RECEIVED, packet=connack, session=self.session
        )
        return connack.return_code

    async def handle_write_timeout(self):
        try:
            if self.session is not None and not self._ping_task:
                self.logger.debug("Scheduling Ping")
                evt = anyio.Event()
                self._tg.start_soon(self.mqtt_ping, evt)
                await evt.wait()
        except BaseException as be:
            self.logger.debug("Exception in ping task: %r", be)
            raise

    async def handle_read_timeout(self):
        pass

    async def mqtt_subscribe(self, topics, packet_id):
        """
        :param topics: array of topics [{'filter':'/a/b', 'qos': 0x00}, ...]
        :return:
        """

        waiter = Future()
        self._subscriptions_waiter[packet_id] = waiter
        try:
            # Build and send SUBSCRIBE message
            subscribe = SubscribePacket.build(topics, packet_id)
            await self._send_packet(subscribe)

            # Wait for SUBACK is received
            return_codes = await waiter.get()

        finally:
            del self._subscriptions_waiter[packet_id]
        return return_codes

    async def handle_suback(self, suback: SubackPacket):
        packet_id = suback.variable_header.packet_id
        try:
            waiter = self._subscriptions_waiter.get(packet_id)
            if waiter is not None:
                await waiter.set(suback.payload.return_codes)
        except KeyError:
            self.logger.warning(
                "Received SUBACK for unknown pending subscription with Id: %s",
                packet_id,
            )

    async def mqtt_unsubscribe(self, topics, packet_id):
        """

        :param topics: array of topics ['/a/b', ...]
        :return:
        """
        waiter = Future()
        self._unsubscriptions_waiter[packet_id] = waiter
        try:
            unsubscribe = UnsubscribePacket.build(topics, packet_id)
            await self._send_packet(unsubscribe)

            await waiter.get()
        finally:
            del self._unsubscriptions_waiter[packet_id]

    async def handle_unsuback(self, unsuback: UnsubackPacket):
        packet_id = unsuback.variable_header.packet_id
        try:
            waiter = self._unsubscriptions_waiter.get(packet_id)
            if waiter is not None:
                await waiter.set(None)
        except KeyError:
            self.logger.warning(
                "Received UNSUBACK for unknown pending subscription with Id: %s",
                packet_id,
            )

    async def mqtt_disconnect(self):
        disconnect_packet = DisconnectPacket()
        await self._send_packet(disconnect_packet)

    async def mqtt_ping(self, evt=None):
        with anyio.CancelScope() as scope:
            self._ping_task = scope
            if evt is not None:
                evt.set()

            try:
                ping_packet = PingReqPacket()
                await self._send_packet(ping_packet)
                resp = await self._pingresp_queue.get()
                return resp
            finally:
                self._ping_task = None

    async def handle_pingresp(self, pingresp: PingRespPacket):
        await self._pingresp_queue.put(pingresp)

    async def handle_connection_closed(self):
        self.logger.debug("Broker closed connection")
        if self._disconnect_waiter:
            self._disconnect_waiter.set()



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import logging
import os
import unittest

import anyio
import pytest

from moat.mqtt.broker import create_broker
from moat.mqtt.client import ConnectException, open_mqttclient
from moat.mqtt.mqtt.constants import QOS_0, QOS_1, QOS_2

from . import anyio_run

formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
logging.basicConfig(level=logging.ERROR, format=formatter)
log = logging.getLogger(__name__)

PORT = 40000 + (os.getpid() + 4) % 10000
URI = "mqtt://127.0.0.1:%d/" % PORT

broker_config = {
    "listeners": {
        "default": {"type": "tcp", "bind": "127.0.0.1:%d" % PORT, "max_connections": 10},
        "ws": {"type": "ws", "bind": "127.0.0.1:8080", "max_connections": 10},
        "wss": {"type": "ws", "bind": "127.0.0.1:8081", "max_connections": 10},
    },
    "sys_interval": 0,
    "auth": {"allow-anonymous": True},
}


class MQTTClientTest(unittest.TestCase):
    def test_connect_tcp(self):
        async def test_coro():
            async with open_mqttclient() as client:
                await client.connect("mqtt://test.mosquitto.org/")
                self.assertIsNotNone(client.session)

        try:
            anyio_run(test_coro)
        except ConnectException:
            log.error("Broken by server")

    def test_connect_tcp_secure(self):
        async def test_coro():
            async with open_mqttclient(config={"check_hostname": False}) as client:
                ca = os.path.join(os.path.dirname(os.path.realpath(__file__)), "mosquitto.org.crt")
                await client.connect("mqtts://test.mosquitto.org/", cafile=ca)
                self.assertIsNotNone(client.session)

        try:
            anyio_run(test_coro)
        except ConnectException:
            log.error("Broken by server")

    def test_connect_tcp_failure(self):
        async def test_coro():
            with pytest.raises(ConnectException):
                config = {"auto_reconnect": False}
                async with open_mqttclient(config=config) as client:
                    await client.connect(URI)

        anyio_run(test_coro)

    def test_uri_supplied_early(self):
        config = {"auto_reconnect": False}

        async def test_coro():
            async with open_mqttclient("mqtt://test.mosquitto.org/", config=config) as client:
                self.assertIsNotNone(client.session)

        try:
            anyio_run(test_coro)
        except ConnectException:
            log.error("Broken by server")

    def test_connect_ws(self):
        async def test_coro():
            async with create_broker(broker_config, plugin_namespace="moat.mqtt.test.plugins"):
                async with open_mqttclient() as client:
                    await client.connect("ws://127.0.0.1:8080/")
                    self.assertIsNotNone(client.session)

        anyio_run(test_coro, backend="trio")

    def test_reconnect_ws_retain_username_password(self):
        async def test_coro():
            async with create_broker(broker_config, plugin_namespace="moat.mqtt.test.plugins"):
                async with open_mqttclient() as client:
                    await client.connect("ws://fred:password@127.0.0.1:8080/")
                    self.assertIsNotNone(client.session)
                    await client.reconnect()

                    self.assertIsNotNone(client.session.username)
                    self.assertIsNotNone(client.session.password)

        anyio_run(test_coro, backend="trio")

    def test_connect_ws_secure(self):
        async def test_coro():
            async with create_broker(broker_config, plugin_namespace="moat.mqtt.test.plugins"):
                async with open_mqttclient() as client:
                    ca = os.path.join(
                        os.path.dirname(os.path.realpath(__file__)), "mosquitto.org.crt"
                    )
                    await client.connect("ws://127.0.0.1:8081/", cafile=ca)
                    self.assertIsNotNone(client.session)

        anyio_run(test_coro, backend="trio")

    def test_ping(self):
        async def test_coro():
            async with create_broker(broker_config, plugin_namespace="moat.mqtt.test.plugins"):
                async with open_mqttclient() as client:
                    await client.connect(URI)
                    self.assertIsNotNone(client.session)
                    await client.ping()

        anyio_run(test_coro, backend="trio")

    def test_subscribe(self):
        async def test_coro():
            async with create_broker(broker_config, plugin_namespace="moat.mqtt.test.plugins"):
                async with open_mqttclient() as client:
                    await client.connect(URI)
                    self.assertIsNotNone(client.session)
                    ret = await client.subscribe(
                        [
                            ("$SYS/broker/uptime", QOS_0),
                            ("$SYS/broker/uptime", QOS_1),
                            ("$SYS/broker/uptime", QOS_2),
                        ]
                    )
                    self.assertEqual(ret[0], QOS_0)
                    self.assertEqual(ret[1], QOS_1)
                    self.assertEqual(ret[2], QOS_2)

        anyio_run(test_coro, backend="trio")

    def test_unsubscribe(self):
        async def test_coro():
            async with create_broker(broker_config, plugin_namespace="moat.mqtt.test.plugins"):
                async with open_mqttclient() as client:
                    await client.connect(URI)
                    self.assertIsNotNone(client.session)
                    ret = await client.subscribe([("$SYS/broker/uptime", QOS_0)])
                    self.assertEqual(ret[0], QOS_0)
                    await client.unsubscribe(["$SYS/broker/uptime"])

        anyio_run(test_coro, backend="trio")

    def test_deliver(self):
        data = b"data"

        async def test_coro():
            async with create_broker(broker_config, plugin_namespace="moat.mqtt.test.plugins"):
                async with open_mqttclient() as client:
                    await client.connect(URI)
                    self.assertIsNotNone(client.session)
                    ret = await client.subscribe([("test_topic", QOS_0)])
                    self.assertEqual(ret[0], QOS_0)
                    async with open_mqttclient() as client_pub:
                        await client_pub.connect(URI)
                        await client_pub.publish("test_topic", data, QOS_0)
                    message = await client.deliver_message()
                    self.assertIsNotNone(message)
                    self.assertIsNotNone(message.publish_packet)
                    self.assertEqual(message.data, data)
                    await client.unsubscribe(["$SYS/broker/uptime"])

        anyio_run(test_coro, backend="trio")

    def test_deliver_timeout(self):
        async def test_coro():
            async with create_broker(broker_config, plugin_namespace="moat.mqtt.test.plugins"):
                async with open_mqttclient() as client:
                    await client.connect(URI)
                    self.assertIsNotNone(client.session)
                    ret = await client.subscribe([("test_topic", QOS_0)])
                    self.assertEqual(ret[0], QOS_0)
                    with self.assertRaises(TimeoutError):
                        with anyio.fail_after(2):
                            await client.deliver_message()
                    await client.unsubscribe(["$SYS/broker/uptime"])

        anyio_run(test_coro, backend="trio")



import pytest
import trio  # noqa: F401
from moat.util import P

from moat.mqtt.test import test_client, test_server


@pytest.mark.trio
async def test_basic(autojump_clock):  # pylint: disable=unused-argument
    async with test_client() as c:
        await c.set(P("test.a.b.c"), value=123)
        res = await c.get(P("test.a.b.c"))
        assert res.value == 123


@pytest.mark.trio
async def test_split(autojump_clock):  # pylint: disable=unused-argument
    async with test_server() as s:
        async with s.test_client() as c1, s.test_client() as c2:
            await c1.set(P("test.a.b.c"), value=123)
            res = await c2.get(P("test.a.b.c"))
            assert res.value == 123



import logging
import anyio

from moat.mqtt.client import open_mqttclient


#
# This sample shows a client running idle.
# Meanwhile, keepalive is managed through PING messages sent every 5 seconds
#


logger = logging.getLogger(__name__)

config = {"keep_alive": 5, "ping_delay": 1}


async def test_coro():
    async with open_mqttclient() as C:
        await C.connect("mqtt://test.mosquitto.org:1883/")
        await anyio.sleep(18)


if __name__ == "__main__":
    formatter = "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.DEBUG, format=formatter)
    anyio.run(test_coro)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import unittest

from moat.mqtt.adapters import BufferAdapter
from moat.mqtt.mqtt.constants import QOS_1, QOS_2
from moat.mqtt.mqtt.packet import PacketIdVariableHeader
from moat.mqtt.mqtt.subscribe import SubscribePacket, SubscribePayload

from .. import anyio_run


class SubscribePacketTest(unittest.TestCase):
    def test_from_stream(self):
        data = b"\x80\x0e\x00\x0a\x00\x03a/b\x01\x00\x03c/d\x02"
        stream = BufferAdapter(data)
        message = anyio_run(SubscribePacket.from_stream, stream)
        (topic, qos) = message.payload.topics[0]
        self.assertEqual(topic, "a/b")
        self.assertEqual(qos, QOS_1)
        (topic, qos) = message.payload.topics[1]
        self.assertEqual(topic, "c/d")
        self.assertEqual(qos, QOS_2)

    def test_to_stream(self):
        variable_header = PacketIdVariableHeader(10)
        payload = SubscribePayload([("a/b", QOS_1), ("c/d", QOS_2)])
        publish = SubscribePacket(variable_header=variable_header, payload=payload)
        out = publish.to_bytes()
        self.assertEqual(out, b"\x82\x0e\x00\x0a\x00\x03a/b\x01\x00\x03c/d\x02")
